import copy

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore import Interaction, JediCoreUtils
from pandajedi.jedicore.InputChunk import InputChunk
from pandajedi.jedicore.MsgWrapper import MsgWrapper

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# class to split job
class JobSplitter:
    # constructor
    def __init__(self):
        self.sizeGradientsPerInSizeForMerge = 1.2
        self.interceptsMerginForMerge = 500 * 1024 * 1024

    # split
    def doSplit(self, taskSpec, inputChunk, siteMapper, allow_chunk_size_limit=False):
        # return for failure
        retFatal = self.SC_FATAL, []
        retTmpError = self.SC_FAILED, []
        # make logger
        tmpLog = MsgWrapper(logger, f"< jediTaskID={taskSpec.jediTaskID} datasetID={inputChunk.masterIndexName} >")
        tmpLog.debug(f"--- start chunk_size_limit={allow_chunk_size_limit}")
        if not inputChunk.isMerging:
            # set maxNumFiles using taskSpec if specified
            maxNumFiles = taskSpec.getMaxNumFilesPerJob()
            # set fsize gradients using taskSpec
            sizeGradients = taskSpec.getOutDiskSize()
            # set fsize intercepts using taskSpec
            sizeIntercepts = taskSpec.getWorkDiskSize()
            # walltime
            if not taskSpec.useHS06():
                walltimeGradient = taskSpec.walltime
            else:
                walltimeGradient = taskSpec.getCpuTime()
            # number of events per job if defined
            nEventsPerJob = taskSpec.getNumEventsPerJob()
            # number of files per job if defined
            nFilesPerJob = taskSpec.getNumFilesPerJob()
            if nFilesPerJob is None and nEventsPerJob is None and inputChunk.useScout() and not taskSpec.useLoadXML() and not taskSpec.respectSplitRule():
                nFilesPerJob = 1
            # grouping with boundaryID
            useBoundary = taskSpec.useGroupWithBoundaryID()
            # fsize intercepts per input size
            sizeGradientsPerInSize = None
            # set max primay output size to avoid producing huge unmerged files
            maxOutSize = 10 * 1024 * 1024 * 1024
            # max size per job
            maxSizePerJob = taskSpec.getMaxSizePerJob()
            if maxSizePerJob is not None:
                maxSizePerJob += InputChunk.defaultOutputSize
            # multiplicity of jobs
            if taskSpec.useJobCloning():
                multiplicity = 1
            else:
                multiplicity = taskSpec.getNumEventServiceConsumer()
            # split with fields
            if taskSpec.getFieldNumToLFN() is not None and taskSpec.useFileAsSourceLFN():
                splitByFields = taskSpec.getFieldNumToLFN()
            else:
                splitByFields = None
        else:
            # set parameters for merging
            maxNumFiles = taskSpec.getMaxNumFilesPerMergeJob()
            sizeGradients = 0
            walltimeGradient = 0
            nFilesPerJob = taskSpec.getNumFilesPerMergeJob()
            nEventsPerJob = taskSpec.getNumEventsPerMergeJob()
            maxSizePerJob = None
            useBoundary = {"inSplit": 3}
            multiplicity = None
            # gradients per input size is 1 + margin
            sizeGradientsPerInSize = self.sizeGradientsPerInSizeForMerge
            # intercepts for libDS
            sizeIntercepts = taskSpec.getWorkDiskSize()
            # mergein of 500MB
            interceptsMergin = self.interceptsMerginForMerge
            if sizeIntercepts < interceptsMergin:
                sizeIntercepts = interceptsMergin
            maxOutSize = taskSpec.getMaxSizePerMergeJob()
            if maxOutSize is None:
                # max output size is 5GB for merging by default
                maxOutSize = 5 * 1024 * 1024 * 1024
            # split with fields
            if taskSpec.getFieldNumToLFN() is not None and taskSpec.useFileAsSourceLFN():
                splitByFields = list(range(4 + 1, 4 + 1 + len(taskSpec.getFieldNumToLFN())))
            else:
                splitByFields = None
        # LB
        respectLB = taskSpec.respectLumiblock()
        # no split for on-site merging
        if taskSpec.on_site_merging():
            no_split = True
        else:
            no_split = False
        # dump
        tmpLog.debug(f"maxNumFiles={maxNumFiles} sizeGradients={sizeGradients} sizeIntercepts={sizeIntercepts} useBoundary={useBoundary}")
        tmpLog.debug(f"walltimeGradient={walltimeGradient} nFilesPerJob={nFilesPerJob} nEventsPerJob={nEventsPerJob}")
        tmpLog.debug(f"useScout={inputChunk.useScout()} isMerging={inputChunk.isMerging}")
        tmpLog.debug(f"sizeGradientsPerInSize={sizeGradientsPerInSize} maxOutSize={maxOutSize} respectLB={respectLB}")
        tmpLog.debug(
            f"multiplicity={multiplicity} splitByFields={str(splitByFields)} "
            f"nFiles={inputChunk.getNumFilesInMaster()} no_split={no_split} "
            f"maxEventsPerJob={taskSpec.get_max_events_per_job()}"
        )
        tmpLog.debug("--- main loop")
        # split
        returnList = []
        subChunks = []
        iSubChunks = 0
        if inputChunk.useScout() and not inputChunk.isMerging:
            default_nSubChunks = 2
        elif taskSpec.is_hpo_workflow():
            default_nSubChunks = 2
        else:
            default_nSubChunks = 25
        subChunk = None
        nSubChunks = default_nSubChunks
        strict_chunkSize = False
        tmp_ng_list = []
        change_site_for_dist_dataset = False
        while True:
            # change site
            if iSubChunks % nSubChunks == 0 or subChunk == [] or change_site_for_dist_dataset:
                change_site_for_dist_dataset = False
                # append to return map
                if subChunks != []:
                    # get site names for parallel execution
                    if taskSpec.getNumSitesPerJob() > 1 and not inputChunk.isMerging and inputChunk.useJumbo != "fake":
                        siteName = inputChunk.getParallelSites(taskSpec.getNumSitesPerJob(), nSubChunks, [siteName])
                    returnList.append(
                        {
                            "siteName": siteName,
                            "subChunks": subChunks,
                            "siteCandidate": siteCandidate,
                        }
                    )
                    try:
                        gshare = taskSpec.gshare.replace(" ", "_")
                    except Exception:
                        gshare = None
                    tmpLog.info(f"split to nJobs={len(subChunks)} at site={siteName} gshare={gshare}")
                    # checkpoint
                    inputChunk.checkpoint_file_usage()
                    # reset
                    subChunks = []
                # skip PQs with chunk size limit
                ngList = copy.copy(tmp_ng_list)
                if not allow_chunk_size_limit:
                    for siteName in inputChunk.get_candidate_names():
                        siteSpec = siteMapper.getSite(siteName)
                        if siteSpec.get_job_chunk_size() is not None:
                            ngList.append(siteName)
                # new candidate
                siteCandidate, getCandidateMsg = inputChunk.getOneSiteCandidate(nSubChunks, ngSites=ngList, get_msg=True)
                if siteCandidate is None:
                    tmpLog.debug(f"no candidate: {getCandidateMsg}")
                    break
                tmp_ng_list = []
                siteName = siteCandidate.siteName
                siteSpec = siteMapper.getSite(siteName)
                # set chunk size
                nSubChunks = siteSpec.get_job_chunk_size()
                if nSubChunks is None:
                    nSubChunks = default_nSubChunks
                    strict_chunkSize = False
                else:
                    strict_chunkSize = True
                # directIO
                if not JediCoreUtils.use_direct_io_for_job(taskSpec, siteSpec, inputChunk):
                    useDirectIO = False
                else:
                    useDirectIO = True
                # get maxSize if it is set in taskSpec
                maxSize = maxSizePerJob
                if maxSize is None:
                    # use maxwdir as the default maxSize
                    if not useDirectIO:
                        maxSize = siteCandidate.get_overridden_attribute("maxwdir")
                        if maxSize is None:
                            maxSize = siteSpec.maxwdir
                        if maxSize:
                            maxSize *= 1024 * 1024
                    elif nEventsPerJob is not None or nFilesPerJob is not None:
                        maxSize = None
                    else:
                        maxSize = siteCandidate.get_overridden_attribute("maxwdir")
                        if maxSize is None:
                            maxSize = siteSpec.maxwdir
                        if inputChunk.useScout():
                            maxSize = max(inputChunk.maxInputSizeScouts, maxSize) * 1024 * 1024
                        else:
                            maxSize = max(inputChunk.maxInputSizeAvalanche, maxSize) * 1024 * 1024
                else:
                    # add offset
                    maxSize += sizeIntercepts
                # max disk size
                maxDiskSize = siteCandidate.get_overridden_attribute("maxwdir")
                if maxDiskSize is None:
                    maxDiskSize = siteSpec.maxwdir
                if maxDiskSize:
                    maxDiskSize *= 1024 * 1024
                # max walltime
                maxWalltime = None
                if not inputChunk.isMerging:
                    maxWalltime = taskSpec.getMaxWalltime()
                if maxWalltime is None:
                    maxWalltime = siteSpec.maxtime
                # core count
                if siteSpec.coreCount:
                    coreCount = siteSpec.coreCount
                else:
                    coreCount = 1
                # core power
                corePower = siteSpec.corepower
                if taskSpec.dynamicNumEvents() and siteSpec.mintime:
                    dynNumEvents = True
                else:
                    dynNumEvents = False
                tmpLog.debug(f"chosen {siteName} : {getCandidateMsg} : nQueue={siteCandidate.nQueuedJobs} nRunCap={siteCandidate.nRunningJobsCap}")
                tmpLog.debug(f"new weight {siteCandidate.weight}")
                tmpLog.debug(
                    f"maxSize={maxSize} maxWalltime={maxWalltime} coreCount={coreCount} corePower={corePower} maxDisk={maxDiskSize} dynNumEvents={dynNumEvents}"
                )
                tmpLog.debug(f"useDirectIO={useDirectIO} label={taskSpec.prodSourceLabel}")
            # get sub chunk
            subChunk = inputChunk.getSubChunk(
                siteName,
                maxSize=maxSize,
                maxNumFiles=maxNumFiles,
                sizeGradients=sizeGradients,
                sizeIntercepts=sizeIntercepts,
                nFilesPerJob=nFilesPerJob,
                walltimeGradient=walltimeGradient,
                maxWalltime=maxWalltime,
                nEventsPerJob=nEventsPerJob,
                useBoundary=useBoundary,
                sizeGradientsPerInSize=sizeGradientsPerInSize,
                maxOutSize=maxOutSize,
                coreCount=coreCount,
                respectLB=respectLB,
                corePower=corePower,
                dynNumEvents=dynNumEvents,
                multiplicity=multiplicity,
                splitByFields=splitByFields,
                tmpLog=tmpLog,
                useDirectIO=useDirectIO,
                maxDiskSize=maxDiskSize,
                enableLog=True,
                no_split=no_split,
                min_walltime=siteSpec.mintime,
                max_events=taskSpec.get_max_events_per_job(),
            )
            if subChunk is None:
                break
            if subChunk != []:
                # append
                subChunks.append(subChunk)
                # check if the last file in master dist dataset is locally available
                for tmp_dataset, tmp_files in subChunk:
                    if tmp_dataset.isMaster():
                        if tmp_dataset.isDistributed() and not siteCandidate.isAvailableFile(tmp_files[-1]) and siteCandidate.isAvailableFile(tmp_files[0]):
                            change_site_for_dist_dataset = True
                            tmpLog.debug(
                                f"change site since the last file in distributed sub-dataset was unavailable at {siteName} while the first file was available"
                            )
                        break
            else:
                tmp_ng_list.append(siteName)
                inputChunk.rollback_file_usage()
            iSubChunks += 1
        # append to return map if remain
        isSkipped = False
        if subChunks != []:
            # skip if chunk size is not enough
            if allow_chunk_size_limit and strict_chunkSize and len(subChunks) < nSubChunks:
                tmpLog.debug(f"skip splitting since chunk size {len(subChunks)} is less than chunk size limit {nSubChunks} at {siteName}")
                inputChunk.rollback_file_usage()
                isSkipped = True
            else:
                # get site names for parallel execution
                if taskSpec.getNumSitesPerJob() > 1 and not inputChunk.isMerging:
                    siteName = inputChunk.getParallelSites(taskSpec.getNumSitesPerJob(), nSubChunks, [siteName])
                returnList.append(
                    {
                        "siteName": siteName,
                        "subChunks": subChunks,
                        "siteCandidate": siteCandidate,
                    }
                )
                try:
                    gshare = taskSpec.gshare.replace(" ", "_")
                except Exception:
                    gshare = None
                tmpLog.info(f"split to nJobs={len(subChunks)} at site={siteName} gshare={gshare}")
        # return
        tmpLog.debug("--- done")
        return self.SC_SUCCEEDED, returnList, isSkipped


Interaction.installSC(JobSplitter)



import copy

from pandajedi.jedicore import JediException
from pandajedi.jedirefine import RefinerUtils

try:
    import idds.common.constants
    import idds.common.utils
    from idds.client.client import Client as iDDS_Client
except ImportError:
    pass


# send notification to external system for additional post-processing
def send_notification(taskBufferIF, ddmIF, taskSpec, tmpLog):
    # send notification to external system
    try:
        taskParam = taskBufferIF.getTaskParamsWithID_JEDI(taskSpec.jediTaskID)
        taskParamMap = RefinerUtils.decodeJSON(taskParam)
    except Exception as e:
        errStr = f"task param conversion from json failed with {str(e)}"
        raise JediException.ExternalTempError(errStr)
    if "outputPostProcessing" in taskParamMap and "system" in taskParamMap["outputPostProcessing"]:
        if taskParamMap["outputPostProcessing"]["system"] == "idds":
            try:
                c = iDDS_Client(idds.common.utils.get_rest_host())
                if taskParamMap["outputPostProcessing"]["type"] == "active_learning":
                    for datasetSpec in taskSpec.datasetSpecList:
                        if datasetSpec.type != "output":
                            continue
                        data = copy.copy(taskParamMap["outputPostProcessing"]["data"])
                        tmp_scope, tmp_name = ddmIF.extract_scope(datasetSpec.datasetName)
                        data["workload_id"] = taskSpec.jediTaskID
                        req = {
                            "scope": tmp_scope,
                            "name": tmp_name,
                            "requester": "panda",
                            "request_type": idds.common.constants.RequestType.ActiveLearning,
                            "transform_tag": idds.common.constants.RequestType.ActiveLearning.value,
                            "status": idds.common.constants.RequestStatus.New,
                            "priority": 0,
                            "lifetime": 30,
                            "request_metadata": data,
                        }
                        tmpLog.debug(f"req {str(req)}")
                        ret = c.add_request(**req)
                        tmpLog.debug(f"got requestID={str(ret)}")
            except Exception as e:
                errStr = f"iDDS failed with {str(e)}"
                raise JediException.ExternalTempError(errStr)



from pandajedi.jedicore import Interaction
from pandajedi.jedicore.MsgWrapper import MsgWrapper

# throttle level
THR_LEVEL5 = 5

LEVEL_None = 0  # There is no configuration defined
LEVEL_GS = 1  # There is a configuration defined at global share level
LEVEL_MS = 2  # There is a configuration defined at MCORE/SCORE level
LEVEL_RT = 3  # There is a configuration defined at resource type level

NQUEUELIMIT = "NQUEUELIMIT"
NRUNNINGCAP = "NRUNNINGCAP"
NQUEUECAP = "NQUEUECAP"

# workqueues that do not work at resource type level.
# E.g. event service is a special case, since MCORE tasks generate SCORE jobs. Therefore we can't work at
# resource type level and need to go to the global level, in order to avoid over-generating jobs
non_rt_wqs = ["eventservice"]


# base class for job throttle
class JobThrottlerBase(object):
    def __init__(self, taskBufferIF):
        self.taskBufferIF = taskBufferIF
        # returns
        self.retTmpError = self.SC_FAILED, True
        self.retThrottled = self.SC_SUCCEEDED, True
        self.retUnThrottled = self.SC_SUCCEEDED, False
        self.retMergeUnThr = self.SC_SUCCEEDED, THR_LEVEL5
        # limit
        self.refresh()
        self.msgType = "jobthrottler"
        self.comp_name = "base_job_throttler"
        self.app = "jedi"

    # refresh
    def refresh(self):
        self.maxNumJobs = None
        self.minPriority = None
        self.underNqLimit = False
        self.siteMapper = self.taskBufferIF.getSiteMapper()

    # set maximum number of jobs to be submitted
    def setMaxNumJobs(self, maxNumJobs):
        self.maxNumJobs = maxNumJobs

    # set min priority of jobs to be submitted
    def setMinPriority(self, minPriority):
        self.minPriority = minPriority

    # check throttle level
    def mergeThrottled(self, thrLevel):
        # un-leveled flag
        if thrLevel in [True, False]:
            return thrLevel
        return thrLevel > THR_LEVEL5

    # check if lack of jobs
    def lackOfJobs(self):
        return self.underNqLimit

    # not enough jobs are queued
    def notEnoughJobsQueued(self):
        self.underNqLimit = True

    def __getConfiguration(self, vo, queue_name, resource_name):
        # component name
        comp_name = self.comp_name
        app = self.app

        # Avoid memory fragmentation
        resource_ms = None
        if resource_name.startswith("MCORE"):
            resource_ms = "MCORE"
        elif resource_name.startswith("SCORE"):
            resource_ms = "SCORE"

        # Read the WQ config values from the DB
        config_map = {
            NQUEUELIMIT: {"value": None, "level": LEVEL_None, "key": None},
            NRUNNINGCAP: {"value": None, "level": LEVEL_None, "key": None},
            NQUEUECAP: {"value": None, "level": LEVEL_None, "key": None},
        }

        for tag in (NQUEUELIMIT, NRUNNINGCAP, NQUEUECAP):
            # 1. try to get a wq + resource_type specific limit
            key_name = f"{tag}_{queue_name}_{resource_name}"
            value = self.taskBufferIF.getConfigValue(comp_name, key_name, app, vo)
            if value:
                config_map[tag] = {"value": value, "level": LEVEL_RT, "key": key_name}
                continue

            # 2. try to get a wq + MCORE/SCORE specific limit
            key_name = f"{tag}_{queue_name}_{resource_ms}*"
            value = self.taskBufferIF.getConfigValue(comp_name, key_name, app, vo)
            if value:
                config_map[tag] = {"value": value, "level": LEVEL_MS, "key": key_name}
                continue

            # 3. try to get a wq specific limit
            key_name = f"{tag}_{queue_name}"
            value = self.taskBufferIF.getConfigValue(comp_name, key_name, app, vo)
            if value:
                config_map[tag] = {"value": value, "level": LEVEL_GS, "key": key_name}

        return config_map

    def __prepareJobStats(self, work_queue, resource_name, config_map):
        """
        Calculates the jobs at resource level (SCORE or MCORE) and in total.

        :param work_queue: work_queue object
        :param resource_name: resource name, e.g. SCORE, MCORE, SCORE_HIMEM, MCORE_HIMEM
        :return: resource_level, nRunning, nRunning_level, nNotRun, nNotRun_level, nDefine, nDefine_level, nWaiting, nWaiting_level
        """
        # SCORE vs MCORE
        if resource_name.startswith("MCORE"):
            ms = "MCORE"
        else:
            ms = "SCORE"

        # get job statistics
        status, wq_stats = self.taskBufferIF.getJobStatisticsByResourceType(work_queue)
        if not status:
            raise RuntimeError("failed to get job statistics")

        # get the number of standby jobs which is used as the number of running jobs
        standby_num_static, standby_num_static_dynamic = self.taskBufferIF.getNumMapForStandbyJobs_JEDI(work_queue)

        # add running if the original stat doesn't have running and standby jobs are required
        if "running" not in wq_stats and (len(standby_num_static) > 0 or len(standby_num_static_dynamic) > 0):
            wq_stats["running"] = {}

        # add dummy to subtract # of starting for dynamic number of standby jobs
        if len(standby_num_static_dynamic) > 0:
            wq_stats["dummy"] = standby_num_static_dynamic

        # Count number of jobs in each status
        # We want to generate one value for the total, one value for the relevant MCORE/SCORE level
        # and one value for the full global share
        nRunning_rt, nRunning_ms, nRunning_gs = 0, 0, 0
        nNotRun_rt, nNotRun_ms, nNotRun_gs = 0, 0, 0
        nDefine_rt, nDefine_ms, nDefine_gs = 0, 0, 0
        nWaiting_rt, nWaiting_gs = 0, 0

        for status in wq_stats:
            nJobs_rt, nJobs_ms, nJobs_gs = 0, 0, 0
            stats_list = list(wq_stats[status].items())
            # take into account the number of standby jobs
            if status == "running":
                stats_list += list(standby_num_static.items())
                stats_list += list(standby_num_static_dynamic.items())
            for resource_type, count in stats_list:
                if resource_type == resource_name:
                    nJobs_rt = count
                if resource_type.startswith(ms):
                    nJobs_ms += count
                nJobs_gs += count

            if status == "running":
                nRunning_rt = nJobs_rt
                nRunning_ms = nJobs_ms
                nRunning_gs = nJobs_gs
            elif status == "defined":
                nDefine_rt = nJobs_rt
                nDefine_ms = nJobs_ms
                nDefine_gs = nJobs_gs
            elif status == "waiting":
                nWaiting_rt = nJobs_rt
                nWaiting_gs = nJobs_gs
            elif status in ["assigned", "activated", "starting"]:
                nNotRun_rt += nJobs_rt
                nNotRun_ms += nJobs_ms
                nNotRun_gs += nJobs_gs
            elif status == "dummy":
                nNotRun_rt -= nJobs_rt
                nNotRun_ms -= nJobs_ms
                nNotRun_gs -= nJobs_gs

        # Get the job stats at the same level as the configured parameters
        # nRunning is compared with the nRunningCap
        if config_map[NRUNNINGCAP]["level"] == LEVEL_GS:
            nRunning_runningcap = nRunning_gs
        elif config_map[NRUNNINGCAP]["level"] == LEVEL_MS:
            nRunning_runningcap = nRunning_ms
        else:
            nRunning_runningcap = nRunning_rt

        # nNotRun and nDefine are compared with the nQueueLimit
        if config_map[NQUEUELIMIT]["level"] == LEVEL_GS:
            nNotRun_queuelimit = nNotRun_gs
            nDefine_queuelimit = nDefine_gs
        elif config_map[NQUEUELIMIT]["level"] == LEVEL_MS:
            nNotRun_queuelimit = nNotRun_ms
            nDefine_queuelimit = nDefine_ms
        else:
            nNotRun_queuelimit = nNotRun_rt
            nDefine_queuelimit = nDefine_rt

        # nNotRun and nDefine are compared with the nQueueCap
        if config_map[NQUEUECAP]["level"] == LEVEL_GS:
            nNotRun_queuecap = nNotRun_gs
            nDefine_queuecap = nDefine_gs
        elif config_map[NQUEUECAP]["level"] == LEVEL_MS:
            nNotRun_queuecap = nNotRun_ms
            nDefine_queuecap = nDefine_ms
        else:
            nNotRun_queuecap = nNotRun_rt
            nDefine_queuecap = nDefine_rt

        return_map = {
            "nRunning_rt": nRunning_rt,
            "nRunning_gs": nRunning_gs,
            "nRunning_runningcap": nRunning_runningcap,
            "nNotRun_rt": nNotRun_rt,
            "nNotRun_gs": nNotRun_gs,
            "nNotRun_queuelimit": nNotRun_queuelimit,
            "nNotRun_queuecap": nNotRun_queuecap,
            "nDefine_rt": nDefine_rt,
            "nDefine_gs": nDefine_gs,
            "nDefine_queuelimit": nDefine_queuelimit,
            "nDefine_queuecap": nDefine_queuecap,
            "nWaiting_rt": nWaiting_rt,
            "nWaiting_gs": nWaiting_gs,
        }

        return return_map

    # check if throttled
    def toBeThrottledBase(self, vo, prodSourceLabel, cloud_name, workQueue, resource_name, logger):
        workqueue_id = workQueue.getID()
        workqueue_name = "_".join(workQueue.queue_name.split(" "))

        # params
        nBunch = 4
        threshold = self.taskBufferIF.getConfigValue(self.comp_name, "THROTTLE_THRESHOLD", self.app, vo)
        if threshold is None:
            threshold = 2.0
        nJobsInBunchMax = 600
        nJobsInBunchMin = 500
        minTotalWalltime = 50 * 1000 * 1000
        nWaitingLimit = 4
        nWaitingBunchLimit = 2
        nParallel = 2
        nParallelCap = 5

        # make logger
        tmp_log = MsgWrapper(logger)
        msg_header = f"{vo}:{prodSourceLabel} cloud={cloud_name} queue={workqueue_name} resource_type={resource_name}:"
        tmp_log.debug(f"{msg_header} start workqueue_id={workqueue_id} threshold={threshold}")

        # check if unthrottled
        if not workQueue.throttled:
            msg_body = "PASS unthrottled since GS_throttled is False"
            tmp_log.info(msg_header + " " + msg_body)
            return self.retUnThrottled

        # get central configuration values
        config_map = self.__getConfiguration(vo, workQueue.queue_name, resource_name)
        configQueueLimit = config_map[NQUEUELIMIT]["value"]
        configQueueLimitKey = config_map[NQUEUELIMIT]["key"]
        configQueueCap = config_map[NQUEUECAP]["value"]
        configQueueCapKey = config_map[NQUEUECAP]["key"]
        configRunningCap = config_map[NRUNNINGCAP]["value"]
        configRunningCapKey = config_map[NRUNNINGCAP]["key"]

        tmp_log.debug(
            msg_header + " got configuration configQueueLimit={} ({}), configQueueCap={} ({}),"
            " configRunningCap={} ({})".format(configQueueLimit, configQueueLimitKey, configQueueCap, configQueueCapKey, configRunningCap, configRunningCapKey)
        )

        # get the jobs statistics for our wq/gs and expand the stats map
        jobstats_map = self.__prepareJobStats(workQueue, resource_name, config_map)

        tmp_log.debug(f"jobstats_map: {jobstats_map}")
        nRunning_rt = jobstats_map["nRunning_rt"]
        nRunning_gs = jobstats_map["nRunning_gs"]
        nRunning_runningcap = jobstats_map["nRunning_runningcap"]
        nNotRun_rt = jobstats_map["nNotRun_rt"]
        nNotRun_gs = jobstats_map["nNotRun_gs"]
        nNotRun_queuelimit = jobstats_map["nNotRun_queuelimit"]
        nNotRun_queuecap = jobstats_map["nNotRun_queuecap"]
        nDefine_rt = jobstats_map["nDefine_rt"]
        nDefine_gs = jobstats_map["nDefine_gs"]
        nDefine_queuelimit = jobstats_map["nDefine_queuelimit"]
        nDefine_queuecap = jobstats_map["nDefine_queuecap"]
        nWaiting_rt = jobstats_map["nWaiting_rt"]
        nWaiting_gs = jobstats_map["nWaiting_gs"]

        # check if higher prio tasks are waiting
        if workQueue.queue_name in non_rt_wqs:
            # find highest priority of currently defined jobs
            tmpStat, highestPrioJobStat = self.taskBufferIF.getHighestPrioJobStat_JEDI(prodSourceLabel, cloud_name, workQueue)
            # the highest priority of waiting tasks
            highestPrioWaiting = self.taskBufferIF.checkWaitingTaskPrio_JEDI(vo, workQueue, prodSourceLabel, cloud_name)
        else:
            # find highest priority of currently defined jobs
            tmpStat, highestPrioJobStat = self.taskBufferIF.getHighestPrioJobStat_JEDI(prodSourceLabel, cloud_name, workQueue, resource_name)
            # the highest priority of waiting tasks
            highestPrioWaiting = self.taskBufferIF.checkWaitingTaskPrio_JEDI(vo, workQueue, prodSourceLabel, cloud_name, resource_name)

        highestPrioInPandaDB = highestPrioJobStat["highestPrio"]
        nNotRunHighestPrio = highestPrioJobStat["nNotRun"]
        if highestPrioWaiting is None:
            msg_body = "failed to get the highest priority of waiting tasks"
            tmp_log.error(f"{msg_header} {msg_body}")
            return self.retTmpError

        # high priority tasks are waiting
        highPrioQueued = False
        if prodSourceLabel != "user":
            if highestPrioWaiting > highestPrioInPandaDB or (highestPrioWaiting == highestPrioInPandaDB and nNotRunHighestPrio < nJobsInBunchMin):
                highPrioQueued = True
        tmp_log.debug(
            "{0} highestPrio waiting:{1} inPanda:{2} numNotRun:{3} -> highPrioQueued={4}".format(
                msg_header, highestPrioWaiting, highestPrioInPandaDB, nNotRunHighestPrio, highPrioQueued
            )
        )
        # set maximum number of jobs to be submitted
        if workQueue.queue_name in non_rt_wqs:
            tmpRemainingSlot = int(nRunning_gs * threshold - nNotRun_gs)
        else:
            tmpRemainingSlot = int(nRunning_rt * threshold - nNotRun_rt)
        # use the lower limit to avoid creating too many _sub/_dis datasets
        nJobsInBunch = min(max(nJobsInBunchMin, tmpRemainingSlot), nJobsInBunchMax)

        if configQueueLimit is not None:
            nQueueLimit = configQueueLimit
        else:
            nQueueLimit = nJobsInBunch * nBunch

        # use nPrestage for reprocessing
        if workQueue.queue_name in ["Heavy Ion", "Reprocessing default"]:
            # reset nJobsInBunch
            if nQueueLimit > (nNotRun_queuelimit + nDefine_queuelimit):
                tmpRemainingSlot = nQueueLimit - (nNotRun_queuelimit + nDefine_queuelimit)
                if tmpRemainingSlot > nJobsInBunch:
                    nJobsInBunch = min(tmpRemainingSlot, nJobsInBunchMax)

        # get cap
        # set number of jobs to be submitted
        if configQueueCap is None:
            self.setMaxNumJobs(nJobsInBunch // nParallel)
        else:
            self.setMaxNumJobs(configQueueCap // nParallelCap)

        # get total walltime
        totWalltime = self.taskBufferIF.getTotalWallTime_JEDI(vo, prodSourceLabel, workQueue, resource_name)

        # log the current situation and limits
        tmp_log.info(f"{msg_header} nQueueLimit={nQueueLimit} nRunCap={configRunningCap} nQueueCap={configQueueCap}")
        tmp_log.info(f"{msg_header} at global share level: nQueued={nNotRun_gs + nDefine_gs} nDefine={nDefine_gs} nRunning={nRunning_gs}")
        tmpMsg = ""
        if config_map[NQUEUECAP]["level"] == LEVEL_MS:
            tmpMsg = f"{msg_header} at MCORE/SCORE level: "
            tmpMsg += f"nQueued_ms={nNotRun_queuecap} "
        if config_map[NRUNNINGCAP]["level"] == LEVEL_MS:
            if not tmpMsg:
                tmpMsg = f"{msg_header} at MCORE/SCORE level: "
            tmpMsg += f"nRunning_ms={nRunning_runningcap} "
        if tmpMsg:
            tmp_log.info(tmpMsg)
        tmp_log.info(
            "{0} at resource type level: nQueued_rt={1} nDefine_rt={2} nRunning_rt={3} totWalltime={4}".format(
                msg_header, nNotRun_rt + nDefine_rt, nDefine_rt, nRunning_rt, totWalltime
            )
        )

        # check number of jobs when high priority jobs are not waiting. test jobs are sent without throttling
        limitPriority = False
        if (
            workQueue.queue_name not in non_rt_wqs
            and nRunning_rt == 0
            and (nNotRun_queuelimit + nDefine_queuelimit) > nQueueLimit
            and (totWalltime is None or totWalltime > minTotalWalltime)
        ):
            limitPriority = True
            if not highPrioQueued:
                # pilot is not running or DDM has a problem
                msg_body = "SKIP no running and enough nQueued_queuelimit={}>({}={}) & totWalltime({})>{} ".format(
                    nNotRun_queuelimit + nDefine_queuelimit, configQueueLimitKey, nQueueLimit, totWalltime, minTotalWalltime
                )
                tmp_log.warning(f"{msg_header} {msg_body}")
                tmp_log.sendMsg(f"{msg_header} {msg_body}", self.msgType, msgLevel="warning", escapeChar=True)
                return self.retMergeUnThr

        elif workQueue.queue_name in non_rt_wqs and nRunning_gs == 0 and (nNotRun_queuelimit + nDefine_queuelimit) > nQueueLimit:
            limitPriority = True
            if not highPrioQueued:
                # pilot is not running or DDM has a problem
                msg_body = f"SKIP no running and enough nQueued_queuelimit={nNotRun_queuelimit + nDefine_queuelimit}>({configQueueLimitKey}={nQueueLimit})"
                tmp_log.warning(f"{msg_header} {msg_body}")
                tmp_log.sendMsg(f"{msg_header} {msg_body}", self.msgType, msgLevel="warning", escapeChar=True)
                return self.retMergeUnThr

        elif (
            workQueue.queue_name not in non_rt_wqs
            and nRunning_rt != 0
            and float(nNotRun_rt + nDefine_rt) / float(nRunning_rt) > threshold
            and (nNotRun_queuelimit + nDefine_queuelimit) > nQueueLimit
            and (totWalltime is None or totWalltime > minTotalWalltime)
        ):
            limitPriority = True
            if not highPrioQueued:
                # enough jobs in Panda
                msg_body = "SKIP nQueued_rt({})/nRunning_rt({})>{} & nQueued_queuelimit={}>({}={}) & totWalltime({})>{}".format(
                    nNotRun_rt + nDefine_rt,
                    nRunning_rt,
                    threshold,
                    nNotRun_queuelimit + nDefine_queuelimit,
                    configQueueLimitKey,
                    nQueueLimit,
                    totWalltime,
                    minTotalWalltime,
                )
                tmp_log.warning(f"{msg_header} {msg_body}")
                tmp_log.sendMsg(f"{msg_header} {msg_body}", self.msgType, msgLevel="warning", escapeChar=True)
                return self.retMergeUnThr

        elif (
            workQueue.queue_name in non_rt_wqs
            and nRunning_gs != 0
            and float(nNotRun_gs + nDefine_gs) / float(nRunning_gs) > threshold
            and (nNotRun_queuelimit + nDefine_queuelimit) > nQueueLimit
        ):
            limitPriority = True
            if not highPrioQueued:
                # enough jobs in Panda
                msg_body = "SKIP nQueued_gs({})/nRunning_gs({})>{} & nQueued_queuelimit={}>({}={})".format(
                    nNotRun_gs + nDefine_gs, nRunning_gs, threshold, nNotRun_queuelimit + nDefine_queuelimit, configQueueLimitKey, nQueueLimit
                )
                tmp_log.warning(f"{msg_header} {msg_body}")
                tmp_log.sendMsg(f"{msg_header} {msg_body}", self.msgType, msgLevel="warning", escapeChar=True)
                return self.retMergeUnThr

        elif nDefine_queuelimit > nQueueLimit:
            limitPriority = True
            if not highPrioQueued:
                # brokerage is stuck
                msg_body = f"SKIP too many nDefined_queuelimit={nDefine_queuelimit}>({configQueueLimitKey}={nQueueLimit})"
                tmp_log.warning(f"{msg_header} {msg_body}")
                tmp_log.sendMsg(f"{msg_header} {msg_body}", self.msgType, msgLevel="warning", escapeChar=True)
                return self.retMergeUnThr

        elif nWaiting_rt > max(nRunning_rt * nWaitingLimit, nJobsInBunch * nWaitingBunchLimit):
            limitPriority = True
            if not highPrioQueued:
                # too many waiting
                msg_body = f"SKIP too many nWaiting_rt({nWaiting_rt})>max(nRunning_rt({nRunning_rt})x{nWaitingLimit},{nJobsInBunch}x{nWaitingBunchLimit})"
                tmp_log.warning(f"{msg_header} {msg_body}")
                tmp_log.sendMsg(f"{msg_header} {msg_body}", self.msgType, msgLevel="warning", escapeChar=True)
                return self.retMergeUnThr

        elif configRunningCap and nRunning_runningcap > configRunningCap:
            # cap on running
            msg_body = f"SKIP nRunning_runningcap({nRunning_runningcap})>nRunningCap({configRunningCapKey}={configRunningCap})"
            tmp_log.warning(f"{msg_header} {msg_body}")
            tmp_log.sendMsg(f"{msg_header} {msg_body}", self.msgType, msgLevel="warning", escapeChar=True)
            return self.retMergeUnThr

        elif configQueueCap and nNotRun_queuecap + nDefine_queuecap > configQueueCap:
            limitPriority = True
            if not highPrioQueued:
                # cap on queued
                msg_body = f"SKIP nQueued_queuecap({nNotRun_queuecap + nDefine_queuecap})>nQueueCap({configQueueCapKey}={configQueueCap})"
                tmp_log.warning(f"{msg_header} {msg_body}")
                tmp_log.sendMsg(f"{msg_header} {msg_body}", self.msgType, msgLevel="warning", escapeChar=True)
                return self.retMergeUnThr

        # get jobs from prodDB
        limitPriorityValue = None
        if limitPriority:
            limitPriorityValue = highestPrioWaiting
            self.setMinPriority(limitPriorityValue)
        else:
            # not enough jobs are queued
            if (
                (nNotRun_queuelimit + nDefine_queuelimit < nQueueLimit * 0.9)
                or (workQueue.queue_name in non_rt_wqs and nNotRun_gs + nDefine_gs < nRunning_gs)
                or (workQueue.queue_name not in non_rt_wqs and nNotRun_rt + nDefine_rt < nRunning_rt)
            ):
                tmp_log.debug(msg_header + " not enough jobs queued")
                self.notEnoughJobsQueued()
                self.setMaxNumJobs(max(self.maxNumJobs, nQueueLimit / 20))

        msg_body = f"PASS - priority limit={limitPriorityValue} maxNumJobs={self.maxNumJobs}"
        tmp_log.info(msg_header + " " + msg_body)
        return self.retUnThrottled


Interaction.installSC(JobThrottlerBase)



import copy
import re
import sys

from pandajedi.jedicore import Interaction
from pandajedi.jedicore.JediDatasetSpec import JediDatasetSpec
from pandajedi.jedicore.JediTaskSpec import JediTaskSpec

from . import RefinerUtils
from .TaskRefinerBase import TaskRefinerBase


# refiner for lost file recovery
class FileRecovery(TaskRefinerBase):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TaskRefinerBase.__init__(self, taskBufferIF, ddmIF)

    # extract common parameters
    def extractCommon(self, jediTaskID, taskParamMap, workQueueMapper):
        return self.SC_SUCCEEDED

    # check matching of dataset names
    def checkDatasetNameMatching(self, datasetName, patternList):
        # name list is not given
        if patternList is None:
            return False
        # loop over all names
        for namePattern in patternList:
            if re.search("^" + namePattern + "$", datasetName) is not None:
                return True
        return False

    # main
    def doRefine(self, jediTaskID, taskParamMap):
        try:
            # make logger
            tmpLog = self.tmpLog
            tmpLog.debug(f"start jediTaskID={jediTaskID}")
            # old dataset name
            oldDatasetName = taskParamMap["oldDatasetName"]
            # accompany datasets
            if "oldAccompanyDatasetNames" in taskParamMap:
                oldAccDatasetNames = taskParamMap["oldAccompanyDatasetNames"]
            else:
                oldAccDatasetNames = None
            # use first file to get task and dataset info
            lostFileName = taskParamMap["lostFiles"][0]
            # get ole jediTaskID and datasetIDs
            tmpStat, oldIDs = self.taskBufferIF.getIDsWithFileDataset_JEDI(oldDatasetName, lostFileName, "output")
            if tmpStat is not True or oldIDs is None:
                tmpLog.error(f"failed to get jediTaskID and DatasetID for {oldDatasetName}:{lostFileName}")
                return self.SC_FAILED
            # get task
            oldJediTaskID = oldIDs["jediTaskID"]
            oldDatasetID = oldIDs["datasetID"]
            tmpStat, oldTaskSpec = self.taskBufferIF.getTaskWithID_JEDI(oldJediTaskID, True)
            if tmpStat is not True:
                tmpLog.error(f"failed to get TaskSpec for old jediTaskId={oldJediTaskID}")
                return self.SC_FAILED
            # make task spec
            taskSpec = JediTaskSpec()
            taskSpec.copyAttributes(oldTaskSpec)
            # reset attributes
            taskSpec.jediTaskID = jediTaskID
            taskSpec.taskType = taskParamMap["taskType"]
            taskSpec.taskPriority = taskParamMap["taskPriority"]
            self.taskSpec = taskSpec
            # get datasets
            tmpStat, datasetSpecList = self.taskBufferIF.getDatasetsWithJediTaskID_JEDI(oldJediTaskID)
            if tmpStat is not True:
                tmpLog.error("failed to get datasetSpecs")
                return self.SC_FAILED
            # loop over all datasets
            provenanceID = None
            dummyStreams = []
            outDatasetSpec = None
            datasetNameSpecMap = {}
            for datasetSpec in datasetSpecList:
                # for output datasets
                if datasetSpec.type not in JediDatasetSpec.getInputTypes():
                    # collect output with the same provenanceID
                    if provenanceID is not None and datasetSpec.provenanceID != provenanceID:
                        continue
                    # set provenanceID if undefined
                    if provenanceID is None and datasetSpec.provenanceID is not None:
                        provenanceID = datasetSpec.provenanceID
                    # collect dummy streams
                    if datasetSpec.type != "log" and (
                        datasetSpec.datasetID != oldDatasetID and not self.checkDatasetNameMatching(datasetSpec.datasetName, oldAccDatasetNames)
                    ):
                        if datasetSpec.streamName not in dummyStreams:
                            dummyStreams.append(datasetSpec.streamName)
                        continue
                # reset attributes
                datasetSpec.status = "defined"
                datasetSpec.datasetID = None
                datasetSpec.jediTaskID = jediTaskID
                datasetSpec.nFiles = 0
                datasetSpec.nFilesUsed = 0
                datasetSpec.nFilesToBeUsed = 0
                datasetSpec.nFilesFinished = 0
                datasetSpec.nFilesFailed = 0
                datasetSpec.nFilesOnHold = 0
                # remove nosplit and repeat since even the same file is made for each bounaryID
                datasetSpec.remNoSplit()
                datasetSpec.remRepeat()
                # append to map
                datasetNameSpecMap[datasetSpec.datasetName] = datasetSpec
                # set master and secondary for input
                if datasetSpec.type in JediDatasetSpec.getInputTypes():
                    if datasetSpec.isMaster():
                        # master
                        self.inMasterDatasetSpec = datasetSpec
                    else:
                        # secondary
                        self.inSecDatasetSpecList.append(datasetSpec)
                elif datasetSpec.type == "log":
                    # set new attributes
                    tmpItem = taskParamMap["log"]
                    datasetSpec.datasetName = tmpItem["dataset"]
                    if "container" in tmpItem:
                        datasetSpec.containerName = tmpItem["container"]
                    if "token" in tmpItem:
                        datasetSpec.storageToken = tmpItem["token"]
                    if "destination" in tmpItem:
                        datasetSpec.destination = tmpItem["destination"]
                    # extract output filename template and change the value field
                    outFileTemplate, tmpItem["value"] = RefinerUtils.extractReplaceOutFileTemplate(tmpItem["value"], datasetSpec.streamName)
                    # make output template
                    if outFileTemplate is not None:
                        if "offset" in tmpItem:
                            offsetVal = 1 + tmpItem["offset"]
                        else:
                            offsetVal = 1
                        outTemplateMap = {
                            "jediTaskID": self.taskSpec.jediTaskID,
                            "serialNr": offsetVal,
                            "streamName": datasetSpec.streamName,
                            "filenameTemplate": outFileTemplate,
                            "outtype": datasetSpec.type,
                        }
                        self.outputTemplateMap[datasetSpec.outputMapKey()] = [outTemplateMap]
                    # append
                    self.outDatasetSpecList.append(datasetSpec)
                else:
                    # output dataset to make copies later
                    outDatasetSpec = datasetSpec
            # replace redundant output streams with dummy files
            for dummyStream in dummyStreams:
                self.taskSpec.jobParamsTemplate = self.taskSpec.jobParamsTemplate.replace("${" + dummyStream + "}", dummyStream.lower() + ".tmp")
            self.setJobParamsTemplate(self.taskSpec.jobParamsTemplate)
            # loop over all lost files
            datasetIDSpecMap = {}
            for lostFileName in taskParamMap["lostFiles"]:
                # get FileID
                tmpStat, tmpIDs = self.taskBufferIF.getIDsWithFileDataset_JEDI(oldDatasetName, lostFileName, "output")
                if tmpStat is not True or tmpIDs is None:
                    tmpLog.error(f"failed to get FileID for {oldDatasetName}:{lostFileName}")
                    return self.SC_FAILED
                # get PandaID
                tmpStat, pandaID = self.taskBufferIF.getPandaIDWithFileID_JEDI(tmpIDs["jediTaskID"], tmpIDs["datasetID"], tmpIDs["fileID"])
                if tmpStat is not True or pandaID is None:
                    tmpLog.error(f"failed to get PandaID for {str(tmpIDs)}")
                    return self.SC_FAILED
                # get files
                tmpStat, fileSpecList = self.taskBufferIF.getFilesWithPandaID_JEDI(pandaID)
                if tmpStat is not True or fileSpecList == []:
                    tmpLog.error(f"failed to get files for PandaID={pandaID}")
                    return self.SC_FAILED
                # append
                for fileSpec in fileSpecList:
                    # only input types
                    if fileSpec.type not in JediDatasetSpec.getInputTypes():
                        continue
                    # get original datasetSpec
                    if fileSpec.datasetID not in datasetIDSpecMap:
                        tmpStat, tmpDatasetSpec = self.taskBufferIF.getDatasetWithID_JEDI(fileSpec.jediTaskID, fileSpec.datasetID)
                        if tmpStat is not True or tmpDatasetSpec is None:
                            tmpLog.error(f"failed to get dataset for jediTaskID={fileSpec.jediTaskID} datasetID={fileSpec.datasetID}")
                            return self.SC_FAILED
                        datasetIDSpecMap[fileSpec.datasetID] = tmpDatasetSpec
                    origDatasetSpec = datasetIDSpecMap[fileSpec.datasetID]
                    if origDatasetSpec.datasetName not in datasetNameSpecMap:
                        tmpLog.error(f"datasetName={origDatasetSpec.datasetName} is missing in new datasets")
                        return self.SC_FAILED
                    # not target or accompany datasets
                    if origDatasetSpec.datasetID != oldDatasetID and not self.checkDatasetNameMatching(origDatasetSpec.datasetName, oldAccDatasetNames):
                        continue
                    newDatasetSpec = datasetNameSpecMap[origDatasetSpec.datasetName]
                    # set new attributes
                    fileSpec.fileID = None
                    fileSpec.datasetID = None
                    fileSpec.jediTaskID = None
                    fileSpec.boundaryID = pandaID
                    fileSpec.keepTrack = 1
                    fileSpec.attemptNr = 1
                    fileSpec.status = "ready"
                    # append
                    newDatasetSpec.addFile(fileSpec)
                # make one output dataset per file
                datasetSpec = copy.copy(outDatasetSpec)
                # set new attributes
                tmpItem = taskParamMap["output"]
                datasetSpec.datasetName = tmpItem["dataset"]
                if "container" in tmpItem:
                    datasetSpec.containerName = tmpItem["container"]
                if "token" in tmpItem:
                    datasetSpec.storageToken = tmpItem["token"]
                if "destination" in tmpItem:
                    datasetSpec.destination = tmpItem["destination"]
                # use PandaID of original job as provenanceID
                datasetSpec.provenanceID = pandaID
                # append
                self.outDatasetSpecList.append(datasetSpec)
                # extract attempt number from original filename
                tmpMatch = re.search("\.(\d+)$", lostFileName)
                if tmpMatch is None:
                    offsetVal = 1
                else:
                    offsetVal = 1 + int(tmpMatch.group(1))
                # filename without attempt number
                baseFileName = re.sub("\.(\d+)$", "", lostFileName)
                # make output template
                outTemplateMap = {
                    "jediTaskID": self.taskSpec.jediTaskID,
                    "serialNr": offsetVal,
                    "streamName": datasetSpec.streamName,
                    "filenameTemplate": baseFileName + ".${SN:d}",
                    "outtype": datasetSpec.type,
                }
                self.outputTemplateMap[datasetSpec.outputMapKey()] = [outTemplateMap]
            # append datasets to task parameters
            for datasetSpec in datasetNameSpecMap.values():
                if datasetSpec.Files == []:
                    continue
                fileList = []
                for fileSpec in datasetSpec.Files:
                    fileList.append(
                        {
                            "lfn": fileSpec.lfn,
                            "firstEvent": fileSpec.firstEvent,
                            "startEvent": fileSpec.startEvent,
                            "endEvent": fileSpec.endEvent,
                            "keepTrack": fileSpec.keepTrack,
                            "boundaryID": fileSpec.boundaryID,
                        }
                    )
                taskParamMap = RefinerUtils.appendDataset(taskParamMap, datasetSpec, fileList)
                self.updatedTaskParams = taskParamMap
            # grouping with boundaryID
            self.setSplitRule(None, 4, JediTaskSpec.splitRuleToken["groupBoundaryID"])
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error(f"doRefine failed with {errtype.__name__}:{errvalue}")
            return self.SC_FAILED
        tmpLog.debug("done")
        return self.SC_SUCCEEDED



# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jediconfig import jedi_config
from pandajedi.jedicore.FactoryBase import FactoryBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# factory class for task setup
class TaskSetupper(FactoryBase):
    # constructor
    def __init__(self, vo, sourceLabel):
        FactoryBase.__init__(self, vo, sourceLabel, logger, jedi_config.tasksetup.modConfig)

    # main
    def doSetup(self, taskSpec, datasetToRegister, pandaJobs):
        return self.getImpl(taskSpec.vo, taskSpec.prodSourceLabel).doSetup(taskSpec, datasetToRegister, pandaJobs)



# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore import Interaction
from pandajedi.jedicore.MsgWrapper import MsgWrapper

from .TaskSetupperBase import TaskSetupperBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# task setup for general purpose
class GenTaskSetupper(TaskSetupperBase):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TaskSetupperBase.__init__(self, taskBufferIF, ddmIF)

    # main to setup task
    def doSetup(self, taskSpec, datasetToRegister, pandaJobs):
        return self.SC_SUCCEEDED



from pandacommon.pandamsgbkr.msg_processor import SimpleMsgProcPluginBase
from pandajedi.jedicore.JediTaskBuffer import JediTaskBuffer


# Base simple message processing plugin
class BaseMsgProcPlugin(SimpleMsgProcPluginBase):
    def initialize(self, in_collective=False):
        """
        initialize plugin instance, run once before loop in thread
        """
        if in_collective:
            # run in collective msg_proc plugin
            pass
        else:
            # run as individual msg_proc plugin
            self.set_tbIF()

    def set_tbIF(self):
        """
        set up JEDI TaskBuffer interface
        """
        # set nDBConnection = n_threads to avoid DBProxy blocking amongs threads
        n_db_conns = getattr(self, "n_threads", 1)
        self.tbIF = JediTaskBuffer(None, nDBConnection=n_db_conns)

    def process(self, msg_obj):
        """
        process the message
        Get msg_obj from the incoming MQ (if any; otherwise msg_obj is None)
        Returned value will be sent to the outgoing MQ (if any)
        """
        pass



# DB API for JEDI

import datetime

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandaserver.brokerage.SiteMapper import SiteMapper
from pandaserver.taskbuffer import TaskBuffer

from pandajedi.jediconfig import jedi_config

from . import JediDBProxyPool
from .Interaction import CommandReceiveInterface

logger = PandaLogger().getLogger(__name__.split(".")[-1])

# use customized proxy pool
TaskBuffer.DBProxyPool = JediDBProxyPool.DBProxyPool


class JediTaskBuffer(TaskBuffer.TaskBuffer, CommandReceiveInterface):
    # constructor
    def __init__(self, conn, nDBConnection=1):
        CommandReceiveInterface.__init__(self, conn)
        TaskBuffer.TaskBuffer.__init__(self)
        TaskBuffer.TaskBuffer.init(self, jedi_config.db.dbhost, jedi_config.db.dbpasswd, nDBConnection=nDBConnection)
        # site mapper
        self.siteMapper = SiteMapper(self)
        # update time for site mapper
        self.dateTimeForSM = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
        logger.debug("__init__")

    # query an SQL
    def querySQL(self, sql, varMap, arraySize=1000):
        with self.proxyPool.get() as proxy:
            return proxy.querySQLS(sql, varMap, arraySize)[1]

    # get SiteMapper
    def getSiteMapper(self):
        timeNow = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
        if datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None) - self.dateTimeForSM > datetime.timedelta(minutes=10):
            self.siteMapper = SiteMapper(self)
            self.dateTimeForSM = timeNow
        return self.siteMapper

    # get work queue map
    def getWorkQueueMap(self):
        with self.proxyPool.get() as proxy:
            return proxy.getWorkQueueMap()

    # get the list of datasets to feed contents to DB
    def getDatasetsToFeedContents_JEDI(self, vo=None, prodSourceLabel=None, task_id=None):
        with self.proxyPool.get() as proxy:
            return proxy.getDatasetsToFeedContents_JEDI(vo, prodSourceLabel, task_id)

    # feed files to the JEDI contents table
    def insertFilesForDataset_JEDI(
        self,
        datasetSpec,
        fileMap,
        datasetState,
        stateUpdateTime,
        nEventsPerFile,
        nEventsPerJob,
        maxAttempt,
        firstEventNumber,
        nMaxFiles,
        nMaxEvents,
        useScout,
        fileList,
        useFilesWithNewAttemptNr,
        nFilesPerJob,
        nEventsPerRange,
        nChunksForScout,
        includePatt,
        excludePatt,
        xmlConfig,
        noWaitParent,
        parent_tid,
        pid,
        maxFailure,
        useRealNumEvents,
        respectLB,
        tgtNumEventsPerJob,
        skipFilesUsedBy,
        ramCount,
        taskSpec,
        skipShortInput,
        inputPreStaging,
        order_by,
        maxFileRecords,
        skip_short_output,
    ):
        with self.proxyPool.get() as proxy:
            return proxy.insertFilesForDataset_JEDI(
                datasetSpec,
                fileMap,
                datasetState,
                stateUpdateTime,
                nEventsPerFile,
                nEventsPerJob,
                maxAttempt,
                firstEventNumber,
                nMaxFiles,
                nMaxEvents,
                useScout,
                fileList,
                useFilesWithNewAttemptNr,
                nFilesPerJob,
                nEventsPerRange,
                nChunksForScout,
                includePatt,
                excludePatt,
                xmlConfig,
                noWaitParent,
                parent_tid,
                pid,
                maxFailure,
                useRealNumEvents,
                respectLB,
                tgtNumEventsPerJob,
                skipFilesUsedBy,
                ramCount,
                taskSpec,
                skipShortInput,
                inputPreStaging,
                order_by,
                maxFileRecords,
                skip_short_output,
            )

    # get files from the JEDI contents table with jediTaskID and/or datasetID
    def getFilesInDatasetWithID_JEDI(self, jediTaskID=None, datasetID=None, nFiles=None, status=None):
        with self.proxyPool.get() as proxy:
            return proxy.getFilesInDatasetWithID_JEDI(jediTaskID, datasetID, nFiles, status)

    # insert dataset to the JEDI datasets table
    def insertDataset_JEDI(self, datasetSpec):
        with self.proxyPool.get() as proxy:
            return proxy.insertDataset_JEDI(datasetSpec)

    # update JEDI dataset
    def updateDataset_JEDI(self, datasetSpec, criteria, lockTask=False):
        with self.proxyPool.get() as proxy:
            return proxy.updateDataset_JEDI(datasetSpec, criteria, lockTask)

    # update JEDI dataset attributes
    def updateDatasetAttributes_JEDI(self, jediTaskID, datasetID, attributes):
        with self.proxyPool.get() as proxy:
            return proxy.updateDatasetAttributes_JEDI(jediTaskID, datasetID, attributes)

    # get JEDI dataset attributes
    def getDatasetAttributes_JEDI(self, jediTaskID, datasetID, attributes):
        with self.proxyPool.get() as proxy:
            return proxy.getDatasetAttributes_JEDI(jediTaskID, datasetID, attributes)

    # get JEDI dataset attributes with map
    def getDatasetAttributesWithMap_JEDI(self, jediTaskID, criteria, attributes):
        with self.proxyPool.get() as proxy:
            return proxy.getDatasetAttributesWithMap_JEDI(jediTaskID, criteria, attributes)

    # get JEDI dataset with jediTaskID and datasetID
    def getDatasetWithID_JEDI(self, jediTaskID, datasetID):
        with self.proxyPool.get() as proxy:
            return proxy.getDatasetWithID_JEDI(jediTaskID, datasetID)

    # get JEDI datasets with jediTaskID
    def getDatasetsWithJediTaskID_JEDI(self, jediTaskID, datasetTypes=None, getFiles=False):
        with self.proxyPool.get() as proxy:
            retStat, datasetSpecList = proxy.getDatasetsWithJediTaskID_JEDI(jediTaskID, datasetTypes=datasetTypes)
            if retStat is True and getFiles is True:
                for datasetSpec in datasetSpecList:
                    # read files
                    retStat, fileSpecList = proxy.getFilesInDatasetWithID_JEDI(jediTaskID, datasetSpec.datasetID, None, None)
                    if retStat is False:
                        break
                    for fileSpec in fileSpecList:
                        datasetSpec.addFile(fileSpec)
            # return
            return retStat, datasetSpecList

    # insert task to the JEDI tasks table
    def insertTask_JEDI(self, taskSpec):
        with self.proxyPool.get() as proxy:
            return proxy.insertTask_JEDI(taskSpec)

    # update JEDI task
    def updateTask_JEDI(self, taskSpec, criteria, oldStatus=None, updateDEFT=False, insertUnknown=None, setFrozenTime=True, setOldModTime=False):
        with self.proxyPool.get() as proxy:
            return proxy.updateTask_JEDI(taskSpec, criteria, oldStatus, updateDEFT, insertUnknown, setFrozenTime, setOldModTime)

    # update JEDI task lock
    def updateTaskLock_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.updateTaskLock_JEDI(jediTaskID)

    # update JEDI task status by ContentsFeeder
    def updateTaskStatusByContFeeder_JEDI(self, jediTaskID, taskSpec=None, getTaskStatus=False, pid=None, setFrozenTime=True, useWorldCloud=False):
        with self.proxyPool.get() as proxy:
            return proxy.updateTaskStatusByContFeeder_JEDI(jediTaskID, taskSpec, getTaskStatus, pid, setFrozenTime, useWorldCloud)

    # get JEDI task with jediTaskID
    def getTaskWithID_JEDI(self, jediTaskID, fullFlag=False, lockTask=False, pid=None, lockInterval=None, clearError=False):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskWithID_JEDI(jediTaskID, fullFlag, lockTask, pid, lockInterval, clearError)

    # get JEDI task and tasks with ID and lock it
    def getTaskDatasetsWithID_JEDI(self, jediTaskID, pid, lockTask=True):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskDatasetsWithID_JEDI(jediTaskID, pid, lockTask)

    # get JEDI tasks with selection criteria
    def getTaskIDsWithCriteria_JEDI(self, criteria, nTasks=50):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskIDsWithCriteria_JEDI(criteria, nTasks)

    # get JEDI tasks to be finished
    def getTasksToBeFinished_JEDI(self, vo, prodSourceLabel, pid, nTasks=50, target_tasks=None):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToBeFinished_JEDI(vo, prodSourceLabel, pid, nTasks, target_tasks)

    # get tasks to be processed
    def getTasksToBeProcessed_JEDI(
        self,
        pid,
        vo,
        workQueue,
        prodSourceLabel,
        cloudName,
        nTasks=50,
        nFiles=100,
        simTasks=None,
        minPriority=None,
        maxNumJobs=None,
        typicalNumFilesMap=None,
        fullSimulation=False,
        simDatasets=None,
        mergeUnThrottled=None,
        readMinFiles=False,
        numNewTaskWithJumbo=0,
        resource_name=None,
        ignore_lock=False,
        target_tasks=None,
    ):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToBeProcessed_JEDI(
                pid,
                vo,
                workQueue,
                prodSourceLabel,
                cloudName,
                nTasks,
                nFiles,
                simTasks=simTasks,
                minPriority=minPriority,
                maxNumJobs=maxNumJobs,
                typicalNumFilesMap=typicalNumFilesMap,
                fullSimulation=fullSimulation,
                simDatasets=simDatasets,
                mergeUnThrottled=mergeUnThrottled,
                readMinFiles=readMinFiles,
                numNewTaskWithJumbo=numNewTaskWithJumbo,
                resource_name=resource_name,
                ignore_lock=ignore_lock,
                target_tasks=target_tasks,
            )

    # get tasks to be processed
    def checkWaitingTaskPrio_JEDI(self, vo, workQueue, prodSourceLabel, cloudName, resource_name=None):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToBeProcessed_JEDI(None, vo, workQueue, prodSourceLabel, cloudName, isPeeking=True, resource_name=resource_name)

    # get job statistics with work queue
    def getJobStatisticsWithWorkQueue_JEDI(self, vo, prodSourceLabel, minPriority=None):
        with self.proxyPool.get() as proxy:
            return proxy.getJobStatisticsWithWorkQueue_JEDI(vo, prodSourceLabel, minPriority)

    # get job statistics by global share
    def getJobStatisticsByGlobalShare(self, vo, exclude_rwq=False):
        with self.proxyPool.get() as proxy:
            return proxy.getJobStatisticsByGlobalShare(vo, exclude_rwq)

    # get job statistics by global share
    def getJobStatisticsByResourceType(self, workqueue):
        with self.proxyPool.get() as proxy:
            return proxy.getJobStatisticsByResourceType(workqueue)

    # get job statistics by site and resource type
    def getJobStatisticsByResourceTypeSite(self, workqueue):
        with self.proxyPool.get() as proxy:
            return proxy.getJobStatisticsByResourceTypeSite(workqueue)

    # generate output files for task
    def getOutputFiles_JEDI(
        self,
        jediTaskID,
        provenanceID,
        simul,
        instantiateTmpl=False,
        instantiatedSite=None,
        isUnMerging=False,
        isPrePro=False,
        xmlConfigJob=None,
        siteDsMap=None,
        middleName="",
        registerDatasets=False,
        parallelOutMap=None,
        fileIDPool=[],
        n_files_per_chunk=1,
    ):
        with self.proxyPool.get() as proxy:
            return proxy.getOutputFiles_JEDI(
                jediTaskID,
                provenanceID,
                simul,
                instantiateTmpl,
                instantiatedSite,
                isUnMerging,
                isPrePro,
                xmlConfigJob,
                siteDsMap,
                middleName,
                registerDatasets,
                parallelOutMap,
                fileIDPool,
                n_files_per_chunk,
            )

    # insert output file templates
    def insertOutputTemplate_JEDI(self, templates):
        with self.proxyPool.get() as proxy:
            return proxy.insertOutputTemplate_JEDI(templates)

    # insert JobParamsTemplate
    def insertJobParamsTemplate_JEDI(self, jediTaskID, templ):
        with self.proxyPool.get() as proxy:
            return proxy.insertJobParamsTemplate_JEDI(jediTaskID, templ)

    # insert TaskParams
    def insertTaskParams_JEDI(self, vo, prodSourceLabel, userName, taskName, taskParams, parent_tid=None):
        with self.proxyPool.get() as proxy:
            return proxy.insertTaskParams_JEDI(vo, prodSourceLabel, userName, taskName, taskParams, parent_tid)

    # reset unused files
    def resetUnusedFiles_JEDI(self, jediTaskID, inputChunk):
        with self.proxyPool.get() as proxy:
            return proxy.resetUnusedFiles_JEDI(jediTaskID, inputChunk)

    # insert TaskParams
    def insertUpdateTaskParams_JEDI(self, jediTaskID, vo, prodSourceLabel, updateTaskParams, insertTaskParamsList):
        with self.proxyPool.get() as proxy:
            return proxy.insertUpdateTaskParams_JEDI(jediTaskID, vo, prodSourceLabel, updateTaskParams, insertTaskParamsList)

    # set missing files
    def setMissingFiles_JEDI(self, jediTaskID, datasetID, fileIDs):
        with self.proxyPool.get() as proxy:
            return proxy.setMissingFiles_JEDI(jediTaskID, datasetID, fileIDs)

    # rescue picked files
    def rescuePickedFiles_JEDI(self, vo, prodSourceLabel, waitTime):
        with self.proxyPool.get() as proxy:
            return proxy.rescuePickedFiles_JEDI(vo, prodSourceLabel, waitTime)

    # rescue unlocked tasks with picked files
    def rescueUnLockedTasksWithPicked_JEDI(self, vo, prodSourceLabel, waitTime, pid):
        with self.proxyPool.get() as proxy:
            return proxy.rescueUnLockedTasksWithPicked_JEDI(vo, prodSourceLabel, waitTime, pid)

    # unlock tasks
    def unlockTasks_JEDI(self, vo, prodSourceLabel, waitTime, hostName=None, pgid=None):
        with self.proxyPool.get() as proxy:
            return proxy.unlockTasks_JEDI(vo, prodSourceLabel, waitTime, hostName, pgid)

    # get the size of input files which will be copied to the site
    def getMovingInputSize_JEDI(self, siteName):
        with self.proxyPool.get() as proxy:
            return proxy.getMovingInputSize_JEDI(siteName)

    # get typical number of input files for each workQueue+processingType
    def getTypicalNumInput_JEDI(self, vo, prodSourceLabel, workQueue):
        with self.proxyPool.get() as proxy:
            return proxy.getTypicalNumInput_JEDI(vo, prodSourceLabel, workQueue)

    # get highest prio jobs with workQueueID
    def getHighestPrioJobStat_JEDI(self, prodSourceLabel, cloudName, workQueue, resource_name=None):
        with self.proxyPool.get() as proxy:
            return proxy.getHighestPrioJobStat_JEDI(prodSourceLabel, cloudName, workQueue, resource_name)

    # get the list of tasks to refine
    def getTasksToRefine_JEDI(self, vo=None, prodSourceLabel=None):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToRefine_JEDI(vo, prodSourceLabel)

    # get task parameters with jediTaskID
    def getTaskParamsWithID_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskParamsWithID_JEDI(jediTaskID)

    # register task/dataset/templ/param in a single transaction
    def registerTaskInOneShot_JEDI(
        self,
        jediTaskID,
        taskSpec,
        inMasterDatasetSpec,
        inSecDatasetSpecList,
        outDatasetSpecList,
        outputTemplateMap,
        jobParamsTemplate,
        taskParams,
        unmergeMasterDatasetSpec,
        unmergeDatasetSpecMap,
        uniqueTaskName,
        oldTaskStatus,
    ):
        with self.proxyPool.get() as proxy:
            return proxy.registerTaskInOneShot_JEDI(
                jediTaskID,
                taskSpec,
                inMasterDatasetSpec,
                inSecDatasetSpecList,
                outDatasetSpecList,
                outputTemplateMap,
                jobParamsTemplate,
                taskParams,
                unmergeMasterDatasetSpec,
                unmergeDatasetSpecMap,
                uniqueTaskName,
                oldTaskStatus,
            )

    # set tasks to be assigned
    def setScoutJobDataToTasks_JEDI(self, vo, prodSourceLabel):
        with self.proxyPool.get() as proxy:
            return proxy.setScoutJobDataToTasks_JEDI(vo, prodSourceLabel, self.siteMapper)

    # prepare tasks to be finished
    def prepareTasksToBeFinished_JEDI(self, vo, prodSourceLabel, nTasks=50, simTasks=None, pid="lock", noBroken=False):
        with self.proxyPool.get() as proxy:
            return proxy.prepareTasksToBeFinished_JEDI(vo, prodSourceLabel, nTasks, simTasks, pid, noBroken, self.siteMapper)

    # get tasks to be assigned
    def getTasksToAssign_JEDI(self, vo, prodSourceLabel, workQueue, resource_name):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToAssign_JEDI(vo, prodSourceLabel, workQueue, resource_name)

    # get tasks to check task assignment
    def getTasksToCheckAssignment_JEDI(self, vo, prodSourceLabel, workQueue, resource_name):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToCheckAssignment_JEDI(vo, prodSourceLabel, workQueue, resource_name)

    # calculate RW with a priority
    def calculateRWwithPrio_JEDI(self, vo, prodSourceLabel, workQueue, priority):
        with self.proxyPool.get() as proxy:
            return proxy.calculateRWwithPrio_JEDI(vo, prodSourceLabel, workQueue, priority)

    # calculate RW for tasks
    def calculateTaskRW_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.calculateTaskRW_JEDI(jediTaskID)

    # calculate WORLD RW with a priority
    def calculateWorldRWwithPrio_JEDI(self, vo, prodSourceLabel, workQueue, priority):
        with self.proxyPool.get() as proxy:
            return proxy.calculateWorldRWwithPrio_JEDI(vo, prodSourceLabel, workQueue, priority)

    # calculate WORLD RW for tasks
    def calculateTaskWorldRW_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.calculateTaskWorldRW_JEDI(jediTaskID)

    # set cloud to tasks
    def setCloudToTasks_JEDI(self, taskCloudMap):
        with self.proxyPool.get() as proxy:
            return proxy.setCloudToTasks_JEDI(taskCloudMap)

    # get the list of tasks to exec command
    def getTasksToExecCommand_JEDI(self, vo, prodSourceLabel):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToExecCommand_JEDI(vo, prodSourceLabel)

    # get the list of PandaIDs for a task
    def getPandaIDsWithTask_JEDI(self, jediTaskID, onlyActive):
        with self.proxyPool.get() as proxy:
            return proxy.getPandaIDsWithTask_JEDI(jediTaskID, onlyActive)

    # get the list of queued PandaIDs for a task
    def getQueuedPandaIDsWithTask_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getQueuedPandaIDsWithTask_JEDI(jediTaskID)

    # get jediTaskID/datasetID/FileID with dataset and file names
    def getIDsWithFileDataset_JEDI(self, datasetName, fileName, fileType):
        with self.proxyPool.get() as proxy:
            return proxy.getIDsWithFileDataset_JEDI(datasetName, fileName, fileType)

    # get PandaID for a file
    def getPandaIDWithFileID_JEDI(self, jediTaskID, datasetID, fileID):
        with self.proxyPool.get() as proxy:
            return proxy.getPandaIDWithFileID_JEDI(jediTaskID, datasetID, fileID)

    # get JEDI files for a job
    def getFilesWithPandaID_JEDI(self, pandaID):
        with self.proxyPool.get() as proxy:
            return proxy.getFilesWithPandaID_JEDI(pandaID)

    # update task parameters
    def updateTaskParams_JEDI(self, jediTaskID, taskParams):
        with self.proxyPool.get() as proxy:
            return proxy.updateTaskParams_JEDI(jediTaskID, taskParams)

    # reactivate pending tasks
    def reactivatePendingTasks_JEDI(self, vo, prodSourceLabel, timeLimit, timeoutLimit=None, minPriority=None):
        with self.proxyPool.get() as proxy:
            return proxy.reactivatePendingTasks_JEDI(vo, prodSourceLabel, timeLimit, timeoutLimit, minPriority)

    # restart contents update
    def restartTasksForContentsUpdate_JEDI(self, vo, prodSourceLabel, timeLimit=30):
        with self.proxyPool.get() as proxy:
            return proxy.restartTasksForContentsUpdate_JEDI(vo, prodSourceLabel, timeLimit=timeLimit)

    # kick exhausted tasks
    def kickExhaustedTasks_JEDI(self, vo, prodSourceLabel, timeLimit):
        with self.proxyPool.get() as proxy:
            return proxy.kickExhaustedTasks_JEDI(vo, prodSourceLabel, timeLimit)

    # get file spec of lib.tgz
    def getBuildFileSpec_JEDI(self, jediTaskID, siteName, associatedSites):
        with self.proxyPool.get() as proxy:
            return proxy.getBuildFileSpec_JEDI(jediTaskID, siteName, associatedSites)

    # get file spec of old lib.tgz
    def getOldBuildFileSpec_JEDI(self, jediTaskID, datasetID, fileID):
        with self.proxyPool.get() as proxy:
            return proxy.getOldBuildFileSpec_JEDI(jediTaskID, datasetID, fileID)

    # insert lib dataset and files
    def insertBuildFileSpec_JEDI(self, jobSpec, reusedDatasetID, simul):
        with self.proxyPool.get() as proxy:
            return proxy.insertBuildFileSpec_JEDI(jobSpec, reusedDatasetID, simul)

    # get sites used by a task
    def getSitesUsedByTask_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getSitesUsedByTask_JEDI(jediTaskID)

    # get random seed
    def getRandomSeed_JEDI(self, jediTaskID, simul):
        with self.proxyPool.get() as proxy:
            return proxy.getRandomSeed_JEDI(jediTaskID, simul)

    # get preprocess metadata
    def getPreprocessMetadata_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getPreprocessMetadata_JEDI(jediTaskID)

    # get log dataset for preprocessing
    def getPreproLog_JEDI(self, jediTaskID, simul):
        with self.proxyPool.get() as proxy:
            return proxy.getPreproLog_JEDI(jediTaskID, simul)

    # get jobsetID
    def getUserJobsetID_JEDI(self, userName):
        with self.proxyPool.get() as proxy:
            tmpJobID, tmpDummy, tmpStat = proxy.getUserParameter(userName, 1, None)
            # return
            return tmpStat, tmpJobID

    # retry or incrementally execute a task
    def retryTask_JEDI(self, jediTaskID, commStr, maxAttempt=5, retryChildTasks=True, discardEvents=False, release_unstaged=False):
        with self.proxyPool.get() as proxy:
            return proxy.retryTask_JEDI(
                jediTaskID, commStr, maxAttempt, retryChildTasks=retryChildTasks, discardEvents=discardEvents, release_unstaged=release_unstaged
            )

    # append input datasets for incremental execution
    def appendDatasets_JEDI(self, jediTaskID, inMasterDatasetSpecList, inSecDatasetSpecList):
        with self.proxyPool.get() as proxy:
            return proxy.appendDatasets_JEDI(jediTaskID, inMasterDatasetSpecList, inSecDatasetSpecList)

    # record retry history
    def recordRetryHistory_JEDI(self, jediTaskID, oldNewPandaIDs, relationType):
        with self.proxyPool.get() as proxy:
            return proxy.recordRetryHistory_JEDI(jediTaskID, oldNewPandaIDs, relationType)

    # get JEDI tasks with a selection criteria
    def getTasksWithCriteria_JEDI(
        self,
        vo,
        prodSourceLabel,
        taskStatusList,
        taskCriteria={},
        datasetCriteria={},
        taskParamList=[],
        datasetParamList=[],
        taskLockColumn=None,
        taskLockInterval=60,
    ):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksWithCriteria_JEDI(
                vo, prodSourceLabel, taskStatusList, taskCriteria, datasetCriteria, taskParamList, datasetParamList, taskLockColumn, taskLockInterval
            )

    # check parent task status
    def checkParentTask_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.checkParentTask_JEDI(jediTaskID)

    # get task status
    def getTaskStatus_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskStatus_JEDI(jediTaskID)

    # get lib.tgz for waiting jobs
    def getLibForWaitingRunJob_JEDI(self, vo, prodSourceLabel, checkInterval):
        with self.proxyPool.get() as proxy:
            return proxy.getLibForWaitingRunJob_JEDI(vo, prodSourceLabel, checkInterval)

    # get tasks to get reassigned
    def getTasksToReassign_JEDI(self, vo=None, prodSourceLabel=None):
        with self.proxyPool.get() as proxy:
            return proxy.getTasksToReassign_JEDI(vo, prodSourceLabel)

    # kill child tasks
    def killChildTasks_JEDI(self, jediTaskID, taskStatus):
        with self.proxyPool.get() as proxy:
            return proxy.killChildTasks_JEDI(jediTaskID, taskStatus)

    # kick child tasks
    def kickChildTasks_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.kickChildTasks_JEDI(jediTaskID)

    # lock task
    def lockTask_JEDI(self, jediTaskID, pid):
        with self.proxyPool.get() as proxy:
            return proxy.lockTask_JEDI(jediTaskID, pid)

    # get successful files
    def getSuccessfulFiles_JEDI(self, jediTaskID, datasetID):
        with self.proxyPool.get() as proxy:
            return proxy.getSuccessfulFiles_JEDI(jediTaskID, datasetID)

    # unlock a single task
    def unlockSingleTask_JEDI(self, jediTaskID, pid):
        with self.proxyPool.get() as proxy:
            return proxy.unlockSingleTask_JEDI(jediTaskID, pid)

    # throttle JEDI tasks
    def throttleTasks_JEDI(self, vo, prodSourceLabel, waitTime):
        with self.proxyPool.get() as proxy:
            return proxy.throttleTasks_JEDI(vo, prodSourceLabel, waitTime)

    # throttle a JEDI task
    def throttleTask_JEDI(self, jediTaskID, waitTime, errorDialog):
        with self.proxyPool.get() as proxy:
            return proxy.throttleTask_JEDI(jediTaskID, waitTime, errorDialog)

    # release throttled tasks
    def releaseThrottledTasks_JEDI(self, vo, prodSourceLabel):
        with self.proxyPool.get() as proxy:
            return proxy.releaseThrottledTasks_JEDI(vo, prodSourceLabel)

    # release throttled task
    def releaseThrottledTask_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.releaseThrottledTask_JEDI(jediTaskID)

    # get throttled users
    def getThrottledUsersTasks_JEDI(self, vo, prodSourceLabel):
        with self.proxyPool.get() as proxy:
            return proxy.getThrottledUsersTasks_JEDI(vo, prodSourceLabel)

    # lock process
    def lockProcess_JEDI(self, vo, prodSourceLabel, cloud, workqueue_id, resource_name, component, pid, forceOption=False, timeLimit=5):
        with self.proxyPool.get() as proxy:
            return proxy.lockProcess_JEDI(vo, prodSourceLabel, cloud, workqueue_id, resource_name, component, pid, forceOption, timeLimit)

    # unlock process
    def unlockProcess_JEDI(self, vo, prodSourceLabel, cloud, workqueue_id, resource_name, component, pid):
        with self.proxyPool.get() as proxy:
            return proxy.unlockProcess_JEDI(vo, prodSourceLabel, cloud, workqueue_id, resource_name, component, pid)

    # unlock process with PID
    def unlockProcessWithPID_JEDI(self, vo, prodSourceLabel, workqueue_id, resource_name, pid, useBase):
        with self.proxyPool.get() as proxy:
            return proxy.unlockProcessWithPID_JEDI(vo, prodSourceLabel, workqueue_id, resource_name, pid, useBase)

    # check process lock
    def checkProcessLock_JEDI(self, vo, prodSourceLabel, cloud, workqueue_id, resource_name, component, pid, checkBase):
        with self.proxyPool.get() as proxy:
            return proxy.checkProcessLock_JEDI(vo, prodSourceLabel, cloud, workqueue_id, resource_name, component, pid, checkBase)

    # get JEDI tasks to be assessed
    def getAchievedTasks_JEDI(self, vo, prodSourceLabel, timeLimit=60, nTasks=50):
        with self.proxyPool.get() as proxy:
            return proxy.getAchievedTasks_JEDI(vo, prodSourceLabel, timeLimit, nTasks)

    # get inactive sites
    def getInactiveSites_JEDI(self, flag, timeLimit):
        with self.proxyPool.get() as proxy:
            return proxy.getInactiveSites_JEDI(flag, timeLimit)

    # get total walltime
    def getTotalWallTime_JEDI(self, vo, prodSourceLabel, workQueue, resource_name):
        with self.proxyPool.get() as proxy:
            return proxy.getTotalWallTime_JEDI(vo, prodSourceLabel, workQueue, resource_name)

    # check duplication with internal merge
    def checkDuplication_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.checkDuplication_JEDI(jediTaskID)

    # get network metrics for brokerage
    def getNetworkMetrics(self, dst, keyList):
        with self.proxyPool.get() as proxy:
            return proxy.getNetworkMetrics(dst, keyList)

    # get nuclei that have built up a long backlog
    def getBackloggedNuclei(self):
        with self.proxyPool.get() as proxy:
            return proxy.getBackloggedNuclei()

    # get network metrics for brokerage
    def getPandaSiteToOutputStorageSiteMapping(self):
        with self.proxyPool.get() as proxy:
            return proxy.getPandaSiteToOutputStorageSiteMapping()

    # get failure counts for a task
    def getFailureCountsForTask_JEDI(self, jediTaskID, timeWindow):
        with self.proxyPool.get() as proxy:
            return proxy.getFailureCountsForTask_JEDI(jediTaskID, timeWindow)

        # count the number of queued jobs per user or working group

    def countJobsPerTarget_JEDI(self, target, is_user):
        with self.proxyPool.get() as proxy:
            return proxy.countJobsPerTarget_JEDI(target, is_user)

    # get old merge job PandaIDs
    def getOldMergeJobPandaIDs_JEDI(self, jediTaskID, pandaID):
        with self.proxyPool.get() as proxy:
            return proxy.getOldMergeJobPandaIDs_JEDI(jediTaskID, pandaID)

    # get active jumbo jobs for a task
    def getActiveJumboJobs_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getActiveJumboJobs_JEDI(jediTaskID)

    # get jobParms of the first job
    def getJobParamsOfFirstJob_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getJobParamsOfFirstJob_JEDI(jediTaskID)

    # bulk fetch fileIDs
    def bulkFetchFileIDs_JEDI(self, jediTaskID, nIDs):
        with self.proxyPool.get() as proxy:
            return proxy.bulkFetchFileIDs_JEDI(jediTaskID, nIDs)

    # set del flag to events
    def setDelFlagToEvents_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.setDelFlagToEvents_JEDI(jediTaskID)

    # set del flag to events
    def removeFilesIndexInconsistent_JEDI(self, jediTaskID, datasetIDs):
        with self.proxyPool.get() as proxy:
            return proxy.removeFilesIndexInconsistent_JEDI(jediTaskID, datasetIDs)

    # throttle jobs in pauses tasks
    def throttleJobsInPausedTasks_JEDI(self, vo, prodSourceLabel):
        with self.proxyPool.get() as proxy:
            return proxy.throttleJobsInPausedTasks_JEDI(vo, prodSourceLabel)

    # set useJumbo flag
    def setUseJumboFlag_JEDI(self, jediTaskID, statusStr):
        with self.proxyPool.get() as proxy:
            return proxy.setUseJumboFlag_JEDI(jediTaskID, statusStr)

    # get number of tasks with running jumbo jobs
    def getNumTasksWithRunningJumbo_JEDI(self, vo, prodSourceLabel, cloudName, workqueue):
        with self.proxyPool.get() as proxy:
            return proxy.getNumTasksWithRunningJumbo_JEDI(vo, prodSourceLabel, cloudName, workqueue)

    # get number of unprocessed events
    def getNumUnprocessedEvents_JEDI(self, vo, prodSourceLabel, criteria, neg_criteria):
        with self.proxyPool.get() as proxy:
            return proxy.getNumUnprocessedEvents_JEDI(vo, prodSourceLabel, criteria, neg_criteria)

    # get number of jobs for a task
    def getNumJobsForTask_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.getNumJobsForTask_JEDI(jediTaskID)

    # get number map for standby jobs
    def getNumMapForStandbyJobs_JEDI(self, workqueue):
        with self.proxyPool.get() as proxy:
            return proxy.getNumMapForStandbyJobs_JEDI(workqueue)

    # update datasets to finish a task
    def updateDatasetsToFinishTask_JEDI(self, jediTaskID, lockedBy):
        with self.proxyPool.get() as proxy:
            return proxy.updateDatasetsToFinishTask_JEDI(jediTaskID, lockedBy)

    # get tasks with jumbo jobs
    def getTaskWithJumbo_JEDI(self, vo, prodSourceLabel):
        with self.proxyPool.get() as proxy:
            return proxy.getTaskWithJumbo_JEDI(vo, prodSourceLabel)

    # kick pending tasks with jumbo jobs
    def kickPendingTasksWithJumbo_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.kickPendingTasksWithJumbo_JEDI(jediTaskID)

    # reset input to re-generate co-jumbo jobs
    def resetInputToReGenCoJumbo_JEDI(self, jediTaskID):
        with self.proxyPool.get() as proxy:
            return proxy.resetInputToReGenCoJumbo_JEDI(jediTaskID)

    # get averaged disk IO
    def getAvgDiskIO_JEDI(self):
        with self.proxyPool.get() as proxy:
            return proxy.getAvgDiskIO_JEDI()

    # update input files stage-in done according to message from idds
    def updateInputFilesStagedAboutIdds_JEDI(self, jeditaskid, scope, filenames_dict):
        with self.proxyPool.get() as proxy:
            return proxy.updateInputFilesStagedAboutIdds_JEDI(jeditaskid, scope, filenames_dict)

    # update input datasets stage-in done according to message from idds
    def updateInputDatasetsStagedAboutIdds_JEDI(self, jeditaskid, scope, dsnames_dict):
        with self.proxyPool.get() as proxy:
            return proxy.updateInputDatasetsStagedAboutIdds_JEDI(jeditaskid, scope, dsnames_dict)

    # get number of staging files
    def getNumStagingFiles_JEDI(self, jeditaskid):
        with self.proxyPool.get() as proxy:
            return proxy.getNumStagingFiles_JEDI(jeditaskid)

    # get usage breakdown by users and sites
    def getUsageBreakdown_JEDI(self, prod_source_label="user"):
        with self.proxyPool.get() as proxy:
            return proxy.getUsageBreakdown_JEDI(prod_source_label)

    # get jobs stat of each user
    def getUsersJobsStats_JEDI(self, prod_source_label="user"):
        with self.proxyPool.get() as proxy:
            return proxy.getUsersJobsStats_JEDI(prod_source_label)

    # insert HPO pseudo event according to message from idds
    def insertHpoEventAboutIdds_JEDI(self, jedi_task_id, event_id_list):
        with self.proxyPool.get() as proxy:
            return proxy.insertHpoEventAboutIdds_JEDI(jedi_task_id, event_id_list)

    # get event statistics
    def get_event_statistics(self, jedi_task_id):
        with self.proxyPool.get() as proxy:
            return proxy.get_event_statistics(jedi_task_id)

    # get site to-running rate statistics by global share
    def getSiteToRunRateStats(self, vo, exclude_rwq, starttime_min, starttime_max):
        with self.proxyPool.get() as proxy:
            return proxy.getSiteToRunRateStats(vo, exclude_rwq, starttime_min, starttime_max)

    # update cache
    def updateCache_JEDI(self, main_key, sub_key, data):
        with self.proxyPool.get() as proxy:
            return proxy.updateCache_JEDI(main_key, sub_key, data)

    # get cache
    def getCache_JEDI(self, main_key, sub_key):
        with self.proxyPool.get() as proxy:
            return proxy.getCache_JEDI(main_key, sub_key)

    # get cache
    def extendSandboxLifetime_JEDI(self, jedi_taskid, file_name):
        with self.proxyPool.get() as proxy:
            return proxy.extendSandboxLifetime_JEDI(jedi_taskid, file_name)

    # turn a task into pending status for some reason
    def makeTaskPending_JEDI(self, jedi_taskid, reason="unknown"):
        with self.proxyPool.get() as proxy:
            return proxy.makeTaskPending_JEDI(jedi_taskid, reason)

    # query tasks and turn them into pending status for some reason, sql_query should query jeditaskid
    def queryTasksToBePending_JEDI(self, sql_query, params_map, reason):
        with self.proxyPool.get() as proxy:
            return proxy.queryTasksToBePending_JEDI(sql_query, params_map, reason)

    # get IDs of all datasets of input and lib, to update data locality records
    def get_tasks_inputdatasets_JEDI(self, vo):
        with self.proxyPool.get() as proxy:
            return proxy.get_tasks_inputdatasets_JEDI(vo)

    # update dataset locality
    def updateDatasetLocality_JEDI(self, jedi_taskid, datasetid, rse):
        with self.proxyPool.get() as proxy:
            return proxy.updateDatasetLocality_JEDI(jedi_taskid, datasetid, rse)

    # delete outdated dataset locality records
    def deleteOutdatedDatasetLocality_JEDI(self, before_timestamp):
        with self.proxyPool.get() as proxy:
            return proxy.deleteOutdatedDatasetLocality_JEDI(before_timestamp)

    # query tasks and preassign them to dedicate workqueue, sql_query should query jeditaskid
    def queryTasksToPreassign_JEDI(self, sql_query, params_map, site, blacklist=set(), limit=1):
        with self.proxyPool.get() as proxy:
            return proxy.queryTasksToPreassign_JEDI(sql_query, params_map, site, blacklist, limit)

    # close and reassign N jobs of a preassigned task
    def reassignJobsInPreassignedTask_JEDI(self, jedi_taskid, site, n_jobs_to_close):
        with self.proxyPool.get() as proxy:
            return proxy.reassignJobsInPreassignedTask_JEDI(jedi_taskid, site, n_jobs_to_close)

    # undo preassigned tasks
    def undoPreassignedTasks_JEDI(self, jedi_taskids, task_orig_attr_map, params_map, force=False):
        with self.proxyPool.get() as proxy:
            return proxy.undoPreassignedTasks_JEDI(jedi_taskids, task_orig_attr_map, params_map, force)

    # set missing files according to iDDS messages
    def setMissingFilesAboutIdds_JEDI(self, jeditaskid, filenames_dict):
        with self.proxyPool.get() as proxy:
            return proxy.setMissingFilesAboutIdds_JEDI(jeditaskid, filenames_dict)

    # set missing files according to iDDS messages
    def load_sw_map(self):
        with self.proxyPool.get() as proxy:
            return proxy.load_sw_map()

    # get origin datasets
    def get_origin_datasets(self, jedi_task_id, dataset_name, lfns):
        with self.proxyPool.get() as proxy:
            return proxy.get_origin_datasets(jedi_task_id, dataset_name, lfns)

    # push message to message processors which triggers functions of agents
    def push_task_trigger_message(self, msg_type, jedi_task_id, data_dict=None, priority=None, task_spec=None):
        with self.proxyPool.get() as proxy:
            return proxy.push_task_trigger_message(msg_type, jedi_task_id, data_dict, priority, task_spec)

    # aggregate carbon footprint of a task
    def get_task_carbon_footprint(self, jedi_task_id, level="global"):
        with self.proxyPool.get() as proxy:
            return proxy.get_task_carbon_footprint(jedi_task_id, level)

    # get pending data carousel tasks and their input datasets
    def get_pending_dc_tasks_JEDI(self, task_type="prod", time_limit_minutes=60):
        with self.proxyPool.get() as proxy:
            return proxy.get_pending_dc_tasks_JEDI(task_type=task_type, time_limit_minutes=time_limit_minutes)

    # get max number of events in a file of the dataset
    def get_max_events_in_dataset(self, jedi_task_id, dataset_id):
        with self.proxyPool.get() as proxy:
            return proxy.get_max_events_in_dataset(jedi_task_id, dataset_id)



"""
Checking DB schema version for PanDA JEDI.
If there is an issue and the pandaEmailNotification var is
defined in panda_config, it will send an email notification.
"""

from packaging import version
from pandajedi.jediconfig import jedi_config
from pandajedi.jedicore import JediDBSchemaInfo
from pandajedi.jedicore.JediDBProxy import DBProxy
from pandaserver.config import panda_config
from pandaserver.srvcore.MailUtils import MailUtils

proxyS = DBProxy()
proxyS.connect(jedi_config.db.dbhost, jedi_config.db.dbpasswd, jedi_config.db.dbuser, jedi_config.db.dbname)

sql = "select major || '.' || minor || '.' || patch from ATLAS_PANDA.pandadb_version where component = 'JEDI'"

res = proxyS.querySQL(sql)
dbVersion = res[0][0]

serverDBVersion = JediDBSchemaInfo.JediDBSchemaInfo().method()

if version.parse(dbVersion) >= version.parse(serverDBVersion):
    print("DB schema check: OK")
else:
    msgBody = (
        "There is an issue with "
        + panda_config.pserveralias
        + ". PanDA DB schema installed is "
        + dbVersion
        + " while PanDA JEDI requires version "
        + serverDBVersion
        + " to be installed. Please check the official docs for instructions on how to upgrade the schema."
    )
    print(msgBody)
    if "pandaEmailNotification" in panda_config.__dict__:
        MailUtils().send(
            panda_config.pandaEmailNotification, "PanDA DB Version installed is not correct for JEDI running on " + panda_config.pserveralias, msgBody
        )



import sys

try:
    metaID = sys.argv[1]
except Exception:
    metaID = None
import json
import uuid

from pandajedi.jedicore.JediTaskBufferInterface import JediTaskBufferInterface

tbIF = JediTaskBufferInterface()
tbIF.setupInterface()

reqIdx = 100000000
dsName = "mc12_8TeV.129933.McAtNloJimmy_AUET2CT10_WW_e_e_0_0_0_0_0.merge.AOD.e1563_s1499_s1504_r3658_r3549_tid00999168_00"
logDatasetName = f"panda.jeditest.log.{uuid.uuid4()}"

taskParamMap = {}

taskParamMap["nFilesPerJob"] = 1
taskParamMap["nFiles"] = 2
taskParamMap["nEventsPerInputFile"] = 10000
taskParamMap["nEventsPerJob"] = 10000
# taskParamMap['nEvents'] = 25000
# taskParamMap['skipScout'] = True
taskParamMap["taskName"] = str(uuid.uuid4())
taskParamMap["userName"] = "pandasrv1"
taskParamMap["vo"] = "atlas"
taskParamMap["taskPriority"] = 900
# taskParamMap['reqID'] = reqIdx
taskParamMap["architecture"] = "i686-slc5-gcc43-opt"
taskParamMap["transUses"] = "Atlas-17.2.2"
taskParamMap["transHome"] = "AtlasProduction-17.2.2.2"
taskParamMap["transPath"] = "Generate_trf.py"
taskParamMap["processingType"] = "evgen"
taskParamMap["prodSourceLabel"] = "test"
taskParamMap["taskType"] = "prod"
taskParamMap["workingGroup"] = "AP_Higgs"
# taskParamMap['coreCount'] = 1
# taskParamMap['walltime'] = 1
taskParamMap["cloud"] = "FR"
# taskParamMap['site'] = 'FZK-LCG2'
taskParamMap["log"] = {
    "dataset": logDatasetName,
    "type": "template",
    "param_type": "log",
    "token": "ATLASDATADISK",
    "offset": 1000,
    "value": f"{logDatasetName}.${{SN}}.log.tgz",
}
outDatasetName = f"panda.jeditest.EVNT.{uuid.uuid4()}"


taskParamMap["jobParameters"] = [
    {
        "type": "template",
        "param_type": "input",
        "value": "inputGenConfFile=${IN}",
        "dataset": dsName,
    },
    {"type": "constant", "value": "ecmEnergy=8000 runNumber=12345"},
    {
        "type": "template",
        "value": "maxEvents=${MAXEVENTS}",
        "param_type": "number",
    },
    {"type": "template", "value": "randomSeed=${RNDMSEED}", "param_type": "number", "offset": 1000},
    {
        "type": "constant",
        "value": "jobConfig=MC12JobOptions/MC12.161559.PowHegPythia8_AU2CT10_ggH145_tautaull.py",
    },
    {
        "type": "template",
        "param_type": "output",
        "token": "ATLASDATADISK",
        "value": f"outputEVNTFile={outDatasetName}.${{SN}}.pool.root",
        "dataset": outDatasetName,
        "offset": 1000,
    },
    {
        "type": "constant",
        "value": "evgenJobOpts=MC12JobOpts-00-01-06_v1.tar.gz",
    },
]

# params to merge output
taskParamMap["mergeOutput"] = True
taskParamMap["mergeSpec"] = {}
taskParamMap["mergeSpec"]["transPath"] = "HITSMerge_tf.py"
taskParamMap["mergeSpec"][
    "jobParameters"
] = "--AMITag s1776 --DBRelease=current --autoConfiguration=everything --outputHitsFile=${OUTPUT0} --inputHitsFile=${TRN_OUTPUT0} --inputLogsFile=${TRN_LOG0}"

jonStr = json.dumps(taskParamMap)

tbIF.insertTaskParams_JEDI(taskParamMap["vo"], taskParamMap["prodSourceLabel"], taskParamMap["userName"], taskParamMap["taskName"], jonStr)



# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore import Interaction
from pandajedi.jedicore.MsgWrapper import MsgWrapper

from .JobThrottlerBase import JobThrottlerBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# class to throttle general jobs
class GenJobThrottler(JobThrottlerBase):
    # constructor
    def __init__(self, taskBufferIF):
        JobThrottlerBase.__init__(self, taskBufferIF)

    # check if throttled
    def toBeThrottled(self, vo, prodSourceLabel, cloudName, workQueue, resourceType):
        # make logger
        tmpLog = MsgWrapper(logger)
        tmpLog.debug(f"start vo={vo} label={prodSourceLabel} cloud={cloudName} workQueue={workQueue.queue_name}")
        # check if unthrottled
        if workQueue.queue_share is None:
            tmpLog.debug("  done : unthrottled since share=None")
            return self.retUnThrottled
        tmpLog.debug("  done : SKIP")
        return self.retThrottled



import os
import re
import socket
import sys
import traceback

# logger
from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandaserver.dataservice.Activator import Activator

from .TypicalWatchDogBase import TypicalWatchDogBase

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# watchdog for ATLAS analysis
class AtlasAnalWatchDog(TypicalWatchDogBase):
    # constructor
    def __init__(self, taskBufferIF, ddmIF):
        TypicalWatchDogBase.__init__(self, taskBufferIF, ddmIF)
        self.pid = f"{socket.getfqdn().split('.')[0]}-{os.getpid()}-dog"
        # self.cronActions = {'forPrestage': 'atlas_prs'}

    # main
    def doAction(self):
        try:
            # get logger
            origTmpLog = MsgWrapper(logger)
            origTmpLog.debug("start")
            # handle waiting jobs
            self.doForWaitingJobs()
            # throttle tasks if so many prestaging requests
            self.doForPreStaging()
            # priority massage
            self.doForPriorityMassage()
            # redo stalled analysis jobs
            self.doForRedoStalledJobs()
            # task share and priority boost
            self.doForTaskBoost()
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            origTmpLog.error(f"failed with {errtype} {errvalue}")
        # return
        origTmpLog.debug("done")
        return self.SC_SUCCEEDED

    # handle waiting jobs
    def doForWaitingJobs(self):
        try:
            tmpLog = MsgWrapper(logger, "doForWaitingJobs label=user")
            # lock
            got_lock = self.taskBufferIF.lockProcess_JEDI(
                vo=self.vo,
                prodSourceLabel=self.prodSourceLabel,
                cloud=None,
                workqueue_id=None,
                resource_name=None,
                component="AtlasAnalWatchDog.doForWaitingJobs",
                pid=self.pid,
                timeLimit=0.5,
            )
            if not got_lock:
                tmpLog.debug("locked by another process. Skipped")
                return
            # check every 60 min
            checkInterval = 60
            # get lib.tgz for waiting jobs
            libList = self.taskBufferIF.getLibForWaitingRunJob_JEDI(self.vo, self.prodSourceLabel, checkInterval)
            tmpLog.debug(f"got {len(libList)} lib.tgz files")
            # activate or kill orphan jobs which were submitted to use lib.tgz when the lib.tgz was being produced
            for prodUserName, datasetName, tmpFileSpec in libList:
                tmpLog = MsgWrapper(logger, f"< #ATM #KV doForWaitingJobs jediTaskID={tmpFileSpec.jediTaskID} label=user >")
                tmpLog.debug("start")
                # check status of lib.tgz
                if tmpFileSpec.status == "failed":
                    # get buildJob
                    pandaJobSpecs = self.taskBufferIF.peekJobs([tmpFileSpec.PandaID], fromDefined=False, fromActive=False, fromWaiting=False)
                    pandaJobSpec = pandaJobSpecs[0]
                    if pandaJobSpec is not None:
                        # kill
                        self.taskBufferIF.updateJobs([pandaJobSpec], False)
                        tmpLog.debug(f'  action=killed_downstream_jobs for user="{prodUserName}" with libDS={datasetName}')
                    else:
                        # PandaJobSpec not found
                        tmpLog.error(f'  cannot find PandaJobSpec for user="{prodUserName}" with PandaID={tmpFileSpec.PandaID}')
                elif tmpFileSpec.status == "finished":
                    # set metadata
                    self.taskBufferIF.setGUIDs(
                        [
                            {
                                "guid": tmpFileSpec.GUID,
                                "lfn": tmpFileSpec.lfn,
                                "checksum": tmpFileSpec.checksum,
                                "fsize": tmpFileSpec.fsize,
                                "scope": tmpFileSpec.scope,
                            }
                        ]
                    )
                    # get lib dataset
                    dataset = self.taskBufferIF.queryDatasetWithMap({"name": datasetName})
                    if dataset is not None:
                        # activate jobs
                        aThr = Activator(self.taskBufferIF, dataset)
                        aThr.start()
                        aThr.join()
                        tmpLog.debug(f'  action=activated_downstream_jobs for user="{prodUserName}" with libDS={datasetName}')
                    else:
                        # datasetSpec not found
                        tmpLog.error(f'  cannot find datasetSpec for user="{prodUserName}" with libDS={datasetName}')
                else:
                    # lib.tgz is not ready
                    tmpLog.debug(f'  keep waiting for user="{prodUserName}" libDS={datasetName}')
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error(f"failed with {errtype} {errvalue} {traceback.format_exc()}")

    # throttle tasks if so many prestaging requests
    def doForPreStaging(self):
        try:
            tmpLog = MsgWrapper(logger, " #ATM #KV doForPreStaging label=user")
            tmpLog.debug("start")
            # lock
            got_lock = self.taskBufferIF.lockProcess_JEDI(
                vo=self.vo,
                prodSourceLabel=self.prodSourceLabel,
                cloud=None,
                workqueue_id=None,
                resource_name=None,
                component="AtlasAnalWatchDog.doForPreStaging",
                pid=self.pid,
                timeLimit=0.5,
            )
            if not got_lock:
                tmpLog.debug("locked by another process. Skipped")
                return
            # get throttled users
            thrUserTasks = self.taskBufferIF.getThrottledUsersTasks_JEDI(self.vo, self.prodSourceLabel)
            # get dispatch datasets
            dispUserTasks = self.taskBufferIF.getDispatchDatasetsPerUser(self.vo, self.prodSourceLabel, True, True)
            # max size of prestaging requests in GB
            maxPrestaging = self.taskBufferIF.getConfigValue("anal_watchdog", "USER_PRESTAGE_LIMIT", "jedi", "atlas")
            if maxPrestaging is None:
                maxPrestaging = 1024
            # max size of transfer requests in GB
            maxTransfer = self.taskBufferIF.getConfigValue("anal_watchdog", "USER_TRANSFER_LIMIT", "jedi", "atlas")
            if maxTransfer is None:
                maxTransfer = 1024
            # throttle interval
            thrInterval = 120
            # loop over all users
            for userName, userDict in dispUserTasks.items():
                # loop over all transfer types
                for transferType, maxSize in [("prestaging", maxPrestaging), ("transfer", maxTransfer)]:
                    if transferType not in userDict:
                        continue
                    userTotal = int(userDict[transferType]["size"] / 1024)
                    tmpLog.debug(f"user={userName} {transferType} total={userTotal} GB")
                    # too large
                    if userTotal > maxSize:
                        tmpLog.debug(f"user={userName} has too large {transferType} total={userTotal} GB > limit={maxSize} GB")
                        # throttle tasks
                        for taskID in userDict[transferType]["tasks"]:
                            if userName not in thrUserTasks or transferType not in thrUserTasks[userName] or taskID not in thrUserTasks[userName][transferType]:
                                tmpLog.debug(f"action=throttle_{transferType} jediTaskID={taskID} for user={userName}")
                                errDiag = f"throttled since transferring large data volume in total={userTotal}GB > limit={maxSize}GB type={transferType}"
                                self.taskBufferIF.throttleTask_JEDI(taskID, thrInterval, errDiag)
                        # remove the user from the list
                        if userName in thrUserTasks and transferType in thrUserTasks[userName]:
                            del thrUserTasks[userName][transferType]
            # release users
            for userName, taskData in thrUserTasks.items():
                for transferType, taskIDs in taskData.items():
                    tmpLog.debug(f"user={userName} release throttled tasks with {transferType}")
                    # unthrottle tasks
                    for taskID in taskIDs:
                        tmpLog.debug(f"action=release_{transferType} jediTaskID={taskID} for user={userName}")
                        self.taskBufferIF.releaseThrottledTask_JEDI(taskID)
            tmpLog.debug("done")
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error(f"failed with {errtype} {errvalue} {traceback.format_exc()}")

    # priority massage
    def doForPriorityMassage(self):
        tmpLog = MsgWrapper(logger, " #ATM #KV doForPriorityMassage label=user")
        tmpLog.debug("start")
        # lock
        got_lock = self.taskBufferIF.lockProcess_JEDI(
            vo=self.vo,
            prodSourceLabel=self.prodSourceLabel,
            cloud=None,
            workqueue_id=None,
            resource_name=None,
            component="AtlasAnalWatchDog.doForPriorityMassage",
            pid=self.pid,
            timeLimit=6,
        )
        if not got_lock:
            tmpLog.debug("locked by another process. Skipped")
            return
        try:
            # get usage breakdown
            usageBreakDownPerUser, usageBreakDownPerSite = self.taskBufferIF.getUsageBreakdown_JEDI(self.prodSourceLabel)
            # get total number of users and running/done jobs
            totalUsers = 0
            totalRunDone = 0
            usersTotalJobs = {}
            usersTotalCores = {}
            for prodUserName in usageBreakDownPerUser:
                wgValMap = usageBreakDownPerUser[prodUserName]
                for workingGroup in wgValMap:
                    siteValMap = wgValMap[workingGroup]
                    totalUsers += 1
                    for computingSite in siteValMap:
                        statValMap = siteValMap[computingSite]
                        totalRunDone += statValMap["rundone"]
                        usersTotalJobs.setdefault(prodUserName, {})
                        usersTotalJobs[prodUserName].setdefault(workingGroup, 0)
                        usersTotalJobs[prodUserName][workingGroup] += statValMap["running"]
                        usersTotalCores.setdefault(prodUserName, {})
                        usersTotalCores[prodUserName].setdefault(workingGroup, 0)
                        usersTotalCores[prodUserName][workingGroup] += statValMap["runcores"]
            tmpLog.debug(f"total {totalUsers} users, {totalRunDone} RunDone jobs")
            # skip if no user
            if totalUsers == 0:
                tmpLog.debug("no user. Skipped...")
                return
            # cap num of running jobs
            tmpLog.debug("cap running jobs")
            prodUserName = None
            maxNumRunPerUser = self.taskBufferIF.getConfigValue("prio_mgr", "CAP_RUNNING_USER_JOBS")
            maxNumRunPerGroup = self.taskBufferIF.getConfigValue("prio_mgr", "CAP_RUNNING_GROUP_JOBS")
            maxNumCorePerUser = self.taskBufferIF.getConfigValue("prio_mgr", "CAP_RUNNING_USER_CORES")
            maxNumCorePerGroup = self.taskBufferIF.getConfigValue("prio_mgr", "CAP_RUNNING_GROUP_CORES")
            if maxNumRunPerUser is None:
                maxNumRunPerUser = 10000
            if maxNumRunPerGroup is None:
                maxNumRunPerGroup = 10000
            if maxNumCorePerUser is None:
                maxNumCorePerUser = 10000
            if maxNumCorePerGroup is None:
                maxNumCorePerGroup = 10000
            try:
                throttledUsers = self.taskBufferIF.getThrottledUsers()
                for prodUserName in usersTotalJobs:
                    for workingGroup in usersTotalJobs[prodUserName]:
                        tmpNumTotalJobs = usersTotalJobs[prodUserName][workingGroup]
                        tmpNumTotalCores = usersTotalCores[prodUserName][workingGroup]
                        if workingGroup is None:
                            maxNumRun = maxNumRunPerUser
                            maxNumCore = maxNumCorePerUser
                        else:
                            maxNumRun = maxNumRunPerGroup
                            maxNumCore = maxNumCorePerGroup
                        if tmpNumTotalJobs >= maxNumRun or tmpNumTotalCores >= maxNumCore:
                            # throttle user
                            tmpNumJobs = self.taskBufferIF.throttleUserJobs(prodUserName, workingGroup, get_dict=True)
                            if tmpNumJobs is not None:
                                for tmpJediTaskID, tmpNumJob in tmpNumJobs.items():
                                    msg = (
                                        'throttled {} jobs in jediTaskID={} for user="{}" group={} ' "since too many running jobs ({} > {}) or cores ({} > {}) "
                                    ).format(tmpNumJob, tmpJediTaskID, prodUserName, workingGroup, tmpNumTotalJobs, maxNumRun, tmpNumTotalCores, maxNumCore)
                                    tmpLog.debug(msg)
                                    tmpLog.sendMsg(msg, "userCap", msgLevel="warning")
                        elif tmpNumTotalJobs < maxNumRun * 0.9 and tmpNumTotalCores < maxNumCore * 0.9 and (prodUserName, workingGroup) in throttledUsers:
                            # unthrottle user
                            tmpNumJobs = self.taskBufferIF.unThrottleUserJobs(prodUserName, workingGroup, get_dict=True)
                            if tmpNumJobs is not None:
                                for tmpJediTaskID, tmpNumJob in tmpNumJobs.items():
                                    msg = (
                                        'released {} jobs in jediTaskID={} for user="{}" group={} '
                                        "since number of running jobs and cores are less than {} and {}"
                                    ).format(tmpNumJob, tmpJediTaskID, prodUserName, workingGroup, maxNumRun, maxNumCore)
                                    tmpLog.debug(msg)
                                    tmpLog.sendMsg(msg, "userCap")
            except Exception as e:
                errStr = f"cap failed for {prodUserName} : {str(e)}"
                errStr.strip()
                errStr += traceback.format_exc()
                tmpLog.error(errStr)
            # to boost
            tmpLog.debug("boost jobs")
            # global average
            globalAverageRunDone = float(totalRunDone) / float(totalUsers)
            tmpLog.debug(f"global average: {globalAverageRunDone}")
            # count the number of users and run/done jobs for each site
            siteRunDone = {}
            siteUsers = {}
            for computingSite in usageBreakDownPerSite:
                userValMap = usageBreakDownPerSite[computingSite]
                for prodUserName in userValMap:
                    wgValMap = userValMap[prodUserName]
                    for workingGroup in wgValMap:
                        statValMap = wgValMap[workingGroup]
                        # count the number of users and running/done jobs
                        siteUsers.setdefault(computingSite, 0)
                        siteUsers[computingSite] += 1
                        siteRunDone.setdefault(computingSite, 0)
                        siteRunDone[computingSite] += statValMap["rundone"]
            # get site average
            tmpLog.debug("site average")
            siteAverageRunDone = {}
            for computingSite in siteRunDone:
                nRunDone = siteRunDone[computingSite]
                siteAverageRunDone[computingSite] = float(nRunDone) / float(siteUsers[computingSite])
                tmpLog.debug(" %-25s : %s" % (computingSite, siteAverageRunDone[computingSite]))
            # check if the number of user's jobs is lower than the average
            for prodUserName in usageBreakDownPerUser:
                wgValMap = usageBreakDownPerUser[prodUserName]
                for workingGroup in wgValMap:
                    tmpLog.debug(f"---> {prodUserName} group={workingGroup}")
                    # count the number of running/done jobs
                    userTotalRunDone = 0
                    for computingSite in wgValMap[workingGroup]:
                        statValMap = wgValMap[workingGroup][computingSite]
                        userTotalRunDone += statValMap["rundone"]
                    # no priority boost when the number of jobs is higher than the average
                    if userTotalRunDone >= globalAverageRunDone:
                        tmpLog.debug(f"enough running {userTotalRunDone} > {globalAverageRunDone} (global average)")
                        continue
                    tmpLog.debug(f"user total:{userTotalRunDone} global average:{globalAverageRunDone}")
                    # check with site average
                    toBeBoostedSites = []
                    for computingSite in wgValMap[workingGroup]:
                        statValMap = wgValMap[workingGroup][computingSite]
                        # the number of running/done jobs is lower than the average and activated jobs are waiting
                        if statValMap["rundone"] >= siteAverageRunDone[computingSite]:
                            tmpLog.debug(f"enough running {statValMap['rundone']} > {siteAverageRunDone[computingSite]} (site average) at {computingSite}")
                        elif statValMap["activated"] == 0:
                            tmpLog.debug(f"no activated jobs at {computingSite}")
                        else:
                            toBeBoostedSites.append(computingSite)
                    # no boost is required
                    if toBeBoostedSites == []:
                        tmpLog.debug("no sites to be boosted")
                        continue
                    # check special prioritized site
                    siteAccessForUser = {}
                    varMap = {}
                    varMap[":dn"] = prodUserName
                    sql = "SELECT pandaSite,pOffset,status,workingGroups FROM ATLAS_PANDAMETA.siteAccess WHERE dn=:dn"
                    res = self.taskBufferIF.querySQL(sql, varMap, arraySize=10000)
                    if res is not None:
                        for pandaSite, pOffset, pStatus, workingGroups in res:
                            # ignore special working group for now
                            if workingGroups not in ["", None]:
                                continue
                            # only approved sites
                            if pStatus != "approved":
                                continue
                            # no priority boost
                            if pOffset == 0:
                                continue
                            # append
                            siteAccessForUser[pandaSite] = pOffset
                    # set weight
                    totalW = 0
                    defaultW = 100
                    for computingSite in toBeBoostedSites:
                        totalW += defaultW
                        if computingSite in siteAccessForUser:
                            totalW += siteAccessForUser[computingSite]
                    totalW = float(totalW)
                    # the total number of jobs to be boosted
                    numBoostedJobs = globalAverageRunDone - float(userTotalRunDone)
                    # get quota
                    quotaFactor = 1.0 + self.taskBufferIF.checkQuota(prodUserName)
                    tmpLog.debug(f"quota factor:{quotaFactor}")
                    # make priority boost
                    nJobsPerPrioUnit = 5
                    highestPrio = 1000
                    for computingSite in toBeBoostedSites:
                        weight = float(defaultW)
                        if computingSite in siteAccessForUser:
                            weight += float(siteAccessForUser[computingSite])
                        weight /= totalW
                        # the number of boosted jobs at the site
                        numBoostedJobsSite = int(numBoostedJobs * weight / quotaFactor)
                        tmpLog.debug(f"nSite:{numBoostedJobsSite} nAll:{numBoostedJobs} W:{weight} Q:{quotaFactor} at {computingSite}")
                        if numBoostedJobsSite / nJobsPerPrioUnit == 0:
                            tmpLog.debug(f"too small number of jobs {numBoostedJobsSite} to be boosted at {computingSite}")
                            continue
                        # get the highest prio of activated jobs at the site
                        varMap = {}
                        varMap[":jobStatus"] = "activated"
                        varMap[":prodSourceLabel"] = self.prodSourceLabel
                        varMap[":pmerge"] = "pmerge"
                        varMap[":prodUserName"] = prodUserName
                        varMap[":computingSite"] = computingSite
                        sql = "SELECT MAX(currentPriority) FROM ATLAS_PANDA.jobsActive4 "
                        sql += "WHERE prodSourceLabel=:prodSourceLabel AND jobStatus=:jobStatus AND computingSite=:computingSite "
                        sql += "AND processingType<>:pmerge AND prodUserName=:prodUserName "
                        if workingGroup is not None:
                            varMap[":workingGroup"] = workingGroup
                            sql += "AND workingGroup=:workingGroup "
                        else:
                            sql += "AND workingGroup IS NULL "
                        res = self.taskBufferIF.querySQL(sql, varMap, arraySize=10)
                        maxPrio = None
                        if res is not None:
                            try:
                                maxPrio = res[0][0]
                            except Exception:
                                pass
                        if maxPrio is None:
                            tmpLog.debug(f"cannot get the highest prio at {computingSite}")
                            continue
                        # delta for priority boost
                        prioDelta = highestPrio - maxPrio
                        # already boosted
                        if prioDelta <= 0:
                            tmpLog.debug(f"already boosted (prio={maxPrio}) at {computingSite}")
                            continue
                        # lower limit
                        minPrio = maxPrio - numBoostedJobsSite / nJobsPerPrioUnit
                        # SQL for priority boost
                        varMap = {}
                        varMap[":jobStatus"] = "activated"
                        varMap[":prodSourceLabel"] = self.prodSourceLabel
                        varMap[":prodUserName"] = prodUserName
                        varMap[":computingSite"] = computingSite
                        varMap[":prioDelta"] = prioDelta
                        varMap[":maxPrio"] = maxPrio
                        varMap[":minPrio"] = minPrio
                        varMap[":rlimit"] = numBoostedJobsSite
                        sql = "UPDATE ATLAS_PANDA.jobsActive4 SET currentPriority=currentPriority+:prioDelta "
                        sql += "WHERE prodSourceLabel=:prodSourceLabel AND prodUserName=:prodUserName "
                        if workingGroup is not None:
                            varMap[":workingGroup"] = workingGroup
                            sql += "AND workingGroup=:workingGroup "
                        else:
                            sql += "AND workingGroup IS NULL "
                        sql += "AND jobStatus=:jobStatus AND computingSite=:computingSite AND currentPriority>:minPrio "
                        sql += "AND currentPriority<=:maxPrio AND rownum<=:rlimit"
                        tmpLog.debug(f"boost {str(varMap)}")
                        res = self.taskBufferIF.querySQL(sql, varMap, arraySize=10)
                        tmpLog.debug(f"   database return : {res}")
            # done
            tmpLog.debug("done")
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error(f"failed with {errtype} {errvalue} {traceback.format_exc()}")

    # redo stalled analysis jobs
    def doForRedoStalledJobs(self):
        tmpLog = MsgWrapper(logger, " #ATM #KV doForRedoStalledJobs label=user")
        tmpLog.debug("start")
        # lock
        got_lock = self.taskBufferIF.lockProcess_JEDI(
            vo=self.vo,
            prodSourceLabel=self.prodSourceLabel,
            cloud=None,
            workqueue_id=None,
            resource_name=None,
            component="AtlasAnalWatchDog.doForRedoStalledJobs",
            pid=self.pid,
            timeLimit=6,
        )
        if not got_lock:
            tmpLog.debug("locked by another process. Skipped")
            return
        # redo stalled analysis jobs
        tmpLog.debug("redo stalled jobs")
        try:
            varMap = {}
            varMap[":prodSourceLabel"] = self.prodSourceLabel
            sqlJ = "SELECT jobDefinitionID,prodUserName FROM ATLAS_PANDA.jobsDefined4 "
            sqlJ += "WHERE prodSourceLabel=:prodSourceLabel AND modificationTime<CURRENT_DATE-2/24 "
            sqlJ += "GROUP BY jobDefinitionID,prodUserName"
            sqlP = "SELECT PandaID FROM ATLAS_PANDA.jobsDefined4 "
            sqlP += "WHERE jobDefinitionID=:jobDefinitionID ANd prodSourceLabel=:prodSourceLabel AND prodUserName=:prodUserName AND rownum <= 1"
            sqlF = "SELECT lfn,type,destinationDBlock FROM ATLAS_PANDA.filesTable4 WHERE PandaID=:PandaID AND status=:status"
            sqlL = "SELECT guid,status,PandaID,dataset FROM ATLAS_PANDA.filesTable4 WHERE lfn=:lfn AND type=:type"
            sqlA = "SELECT PandaID FROM ATLAS_PANDA.jobsDefined4 "
            sqlA += "WHERE jobDefinitionID=:jobDefinitionID ANd prodSourceLabel=:prodSourceLabel AND prodUserName=:prodUserName"
            sqlU = "UPDATE ATLAS_PANDA.jobsDefined4 SET modificationTime=CURRENT_DATE "
            sqlU += "WHERE jobDefinitionID=:jobDefinitionID ANd prodSourceLabel=:prodSourceLabel AND prodUserName=:prodUserName"
            # get stalled jobs
            resJ = self.taskBufferIF.querySQL(sqlJ, varMap)
            if resJ is None or len(resJ) == 0:
                pass
            else:
                # loop over all jobID/users
                for jobDefinitionID, prodUserName in resJ:
                    tmpLog.debug(f" user:{prodUserName} jobID:{jobDefinitionID}")
                    # get stalled jobs
                    varMap = {}
                    varMap[":prodSourceLabel"] = self.prodSourceLabel
                    varMap[":jobDefinitionID"] = jobDefinitionID
                    varMap[":prodUserName"] = prodUserName
                    resP = self.taskBufferIF.querySQL(sqlP, varMap)
                    if resP is None or len(resP) == 0:
                        tmpLog.debug("  no PandaID")
                        continue
                    useLib = False
                    libStatus = None
                    libGUID = None
                    libLFN = None
                    libDSName = None
                    destReady = False
                    # use the first PandaID
                    for (PandaID,) in resP:
                        tmpLog.debug(f"  check PandaID:{PandaID}")
                        # get files
                        varMap = {}
                        varMap[":PandaID"] = PandaID
                        varMap[":status"] = "unknown"
                        resF = self.taskBufferIF.querySQL(sqlF, varMap)
                        if resF is None or len(resF) == 0:
                            tmpLog.debug("  no files")
                        else:
                            # get lib.tgz and destDBlock
                            for lfn, filetype, destinationDBlock in resF:
                                if filetype == "input" and lfn.endswith(".lib.tgz"):
                                    useLib = True
                                    libLFN = lfn
                                    varMap = {}
                                    varMap[":lfn"] = lfn
                                    varMap[":type"] = "output"
                                    resL = self.taskBufferIF.querySQL(sqlL, varMap)
                                    # not found
                                    if resL is None or len(resL) == 0:
                                        tmpLog.error(f"  cannot find status of {lfn}")
                                        continue
                                    # check status
                                    guid, outFileStatus, pandaIDOutLibTgz, tmpLibDsName = resL[0]
                                    tmpLog.debug(f"  PandaID:{pandaIDOutLibTgz} produces {tmpLibDsName}:{lfn} GUID={guid} status={outFileStatus}")
                                    libStatus = outFileStatus
                                    libGUID = guid
                                    libDSName = tmpLibDsName
                                elif filetype in ["log", "output"]:
                                    if destinationDBlock is not None and re.search("_sub\d+$", destinationDBlock) is not None:
                                        destReady = True
                            break
                    tmpLog.debug(f"  useLib:{useLib} libStatus:{libStatus} libDsName:{libDSName} libLFN:{libLFN} libGUID:{libGUID} destReady:{destReady}")
                    if libStatus == "failed":
                        # delete downstream jobs
                        tmpLog.debug("  -> delete downstream jobs")
                        # FIXME
                        # self.taskBufferIF.deleteStalledJobs(libLFN)
                    else:
                        # activate
                        if useLib and libStatus == "ready" and (libGUID not in [None, ""]) and (libDSName not in [None, ""]):
                            # update GUID
                            tmpLog.debug(f"  set GUID:{libGUID} for {libLFN}")
                            # retG = self.taskBufferIF.setGUIDs([{'lfn':libLFN,'guid':libGUID}])
                            # FIXME
                            retG = True
                            if not retG:
                                tmpLog.error(f"  failed to update GUID for {libLFN}")
                            else:
                                # get PandaID with lib.tgz
                                # ids = self.taskBufferIF.updateInFilesReturnPandaIDs(libDSName,'ready')
                                ids = []
                                # get jobs
                                jobs = self.taskBufferIF.peekJobs(ids, fromActive=False, fromArchived=False, fromWaiting=False)
                                # remove None and unknown
                                acJobs = []
                                for job in jobs:
                                    if job is None or job.jobStatus == "unknown":
                                        continue
                                    acJobs.append(job)
                                # activate
                                tmpLog.debug("  -> activate downstream jobs")
                                # self.taskBufferIF.activateJobs(acJobs)
                        else:
                            # wait
                            tmpLog.debug("  -> wait")
                            varMap = {}
                            varMap[":prodSourceLabel"] = self.prodSourceLabel
                            varMap[":jobDefinitionID"] = jobDefinitionID
                            varMap[":prodUserName"] = prodUserName
                            # FIXME
                            # resU = self.taskBufferIF.querySQL(sqlU, varMap)
            # done
            tmpLog.debug("done")
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error(f"failed to redo stalled jobs with {errtype} {errvalue} {traceback.format_exc()}")

    # task share and priority boost
    def doForTaskBoost(self):
        tmpLog = MsgWrapper(logger, " #ATM #KV doForTaskBoost label=user")
        tmpLog.debug("start")
        # lock
        got_lock = self.taskBufferIF.lockProcess_JEDI(
            vo=self.vo,
            prodSourceLabel=self.prodSourceLabel,
            cloud=None,
            workqueue_id=None,
            resource_name=None,
            component="AtlasAnalWatchDog.doForTaskBoost",
            pid=self.pid,
            timeLimit=5,
        )
        if not got_lock:
            tmpLog.debug("locked by another process. Skipped")
            return
        try:
            # get active tasks in S-class
            sql_get_tasks = (
                """SELECT tev.value_json.task_id, tev.value_json."user" """
                """FROM ATLAS_PANDA.Task_Evaluation tev, ATLAS_PANDA.JEDI_Tasks t """
                """WHERE tev.value_json.task_id=t.jediTaskID """
                """AND tev.metric='analy_task_eval' """
                """AND tev.value_json.class=:s_class """
                """AND tev.value_json.gshare=:gshare """
                """AND t.gshare=:gshare """
            )
            # varMap
            varMap = {
                ":s_class": 2,
                ":gshare": "User Analysis",
            }
            # result
            res = self.taskBufferIF.querySQL(sql_get_tasks, varMap)
            #  Assign to Express Analysis
            new_share = "Express Analysis"
            for taskID, user in res:
                if False:
                    # dry-run
                    tmpLog.debug(
                        "(dry-run) action=gshare_reassignment jediTaskID={} user={} from gshare_old={} to gshare_new={} label=user".format(
                            taskID, user, varMap[":gshare"], new_share
                        )
                    )
                else:
                    tmpLog.info(
                        f" >>> action=gshare_reassignment jediTaskID={taskID} from gshare_old={varMap[':gshare']} to gshare_new={new_share} #ATM #KV label=user"
                    )
                    self.taskBufferIF.reassignShare([taskID], new_share, True)
                    tmpLog.info(f">>> done jediTaskID={taskID}")
            # done
            tmpLog.debug("done")
        except Exception:
            errtype, errvalue = sys.exc_info()[:2]
            tmpLog.error(f"failed with {errtype} {errvalue} {traceback.format_exc()}")



import datetime
import os
import re
import socket
import time
import traceback

from pandacommon.pandalogger.PandaLogger import PandaLogger
from pandajedi.jediconfig import jedi_config
from pandajedi.jedicore import Interaction
from pandajedi.jedicore.JediTaskSpec import JediTaskSpec
from pandajedi.jedicore.MsgWrapper import MsgWrapper
from pandajedi.jedicore.ThreadUtils import ListWithLock, ThreadPool, WorkerThread
from pandajedi.jedirefine import RefinerUtils

from .JediKnight import JediKnight

logger = PandaLogger().getLogger(__name__.split(".")[-1])


# worker class to kill/amend tasks
class TaskCommando(JediKnight):
    # constructor
    def __init__(self, commuChannel, taskBufferIF, ddmIF, vos, prodSourceLabels):
        self.vos = self.parseInit(vos)
        self.prodSourceLabels = self.parseInit(prodSourceLabels)
        self.pid = f"{socket.getfqdn().split('.')[0]}-{os.getpid()}-dog"
        JediKnight.__init__(self, commuChannel, taskBufferIF, ddmIF, logger)

    # main
    def start(self):
        # start base classes
        JediKnight.start(self)
        # go into main loop
        while True:
            startTime = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None)
            try:
                # get logger
                tmpLog = MsgWrapper(logger)
                tmpLog.debug("start")
                # loop over all vos
                for vo in self.vos:
                    # loop over all sourceLabels
                    for prodSourceLabel in self.prodSourceLabels:
                        # get the list of tasks to exec command
                        tmpList = self.taskBufferIF.getTasksToExecCommand_JEDI(vo, prodSourceLabel)
                        if tmpList is None:
                            # failed
                            tmpLog.error(f"failed to get the task list for vo={vo} label={prodSourceLabel}")
                        else:
                            tmpLog.debug(f"got {len(tmpList)} tasks")
                            # put to a locked list
                            taskList = ListWithLock(tmpList)
                            # make thread pool
                            threadPool = ThreadPool()
                            # make workers
                            nWorker = jedi_config.taskrefine.nWorkers
                            for iWorker in range(nWorker):
                                thr = TaskCommandoThread(taskList, threadPool, self.taskBufferIF, self.ddmIF, self.pid)
                                thr.start()
                            # join
                            threadPool.join()
                tmpLog.debug("done")
            except Exception as e:
                tmpLog.error(f"failed in {self.__class__.__name__}.start() with {str(e)} {traceback.format_exc()}")
            # sleep if needed
            loopCycle = jedi_config.tcommando.loopCycle
            timeDelta = datetime.datetime.now(datetime.timezone.utc).replace(tzinfo=None) - startTime
            sleepPeriod = loopCycle - timeDelta.seconds
            if sleepPeriod > 0:
                time.sleep(sleepPeriod)
            # randomize cycle
            self.randomSleep(max_val=loopCycle)


# thread for real worker
class TaskCommandoThread(WorkerThread):
    # constructor
    def __init__(self, taskList, threadPool, taskbufferIF, ddmIF, pid):
        # initialize woker with no semaphore
        WorkerThread.__init__(self, None, threadPool, logger)
        # attributres
        self.taskList = taskList
        self.taskBufferIF = taskbufferIF
        self.ddmIF = ddmIF
        self.msgType = "taskcommando"
        self.pid = pid

    # main
    def runImpl(self):
        while True:
            try:
                # get a part of list
                nTasks = 10
                taskList = self.taskList.get(nTasks)
                # no more datasets
                if len(taskList) == 0:
                    self.logger.debug(f"{self.__class__.__name__} terminating since no more items")
                    return
                # loop over all tasks
                for jediTaskID, commandMap in taskList:
                    # make logger
                    tmpLog = MsgWrapper(self.logger, f" < jediTaskID={jediTaskID} >")
                    commandStr = commandMap["command"]
                    commentStr = commandMap["comment"]
                    oldStatus = commandMap["oldStatus"]
                    tmpLog.info(f"start for {commandStr}")
                    tmpStat = Interaction.SC_SUCCEEDED
                    if commandStr in ["kill", "finish", "reassign"]:
                        tmpMsg = f"executing {commandStr}"
                        tmpLog.info(tmpMsg)
                        tmpLog.sendMsg(tmpMsg, self.msgType)
                        # loop twice to see immediate result
                        for iLoop in range(2):
                            # get active PandaIDs to be killed
                            if commandStr == "reassign" and commentStr is not None and "soft reassign" in commentStr:
                                pandaIDs = self.taskBufferIF.getQueuedPandaIDsWithTask_JEDI(jediTaskID)
                            elif commandStr == "reassign" and commentStr is not None and "nokill reassign" in commentStr:
                                pandaIDs = []
                            else:
                                pandaIDs = self.taskBufferIF.getPandaIDsWithTask_JEDI(jediTaskID, True)
                            if pandaIDs is None:
                                tmpLog.error(f"failed to get PandaIDs for jediTaskID={jediTaskID}")
                                tmpStat = Interaction.SC_FAILED
                            # kill jobs or update task
                            if tmpStat == Interaction.SC_SUCCEEDED:
                                if pandaIDs == []:
                                    # done since no active jobs
                                    tmpMsg = "completed cleaning jobs"
                                    tmpLog.sendMsg(tmpMsg, self.msgType)
                                    tmpLog.info(tmpMsg)
                                    tmpTaskSpec = JediTaskSpec()
                                    tmpTaskSpec.jediTaskID = jediTaskID
                                    updateTaskStatus = True
                                    if commandStr != "reassign":
                                        # reset oldStatus
                                        # keep oldStatus for task reassignment since it is reset when actually reassigned
                                        tmpTaskSpec.forceUpdate("oldStatus")
                                    else:
                                        # extract cloud or site
                                        if commentStr is not None:
                                            tmpItems = commentStr.split(":")
                                            if tmpItems[0] == "cloud":
                                                tmpTaskSpec.cloud = tmpItems[1]
                                            elif tmpItems[0] == "nucleus":
                                                tmpTaskSpec.nucleus = tmpItems[1]
                                            else:
                                                tmpTaskSpec.site = tmpItems[1]
                                            tmpMsg = f"set {tmpItems[0]}={tmpItems[1]}"
                                            tmpLog.sendMsg(tmpMsg, self.msgType)
                                            tmpLog.info(tmpMsg)
                                            # back to oldStatus if necessary
                                            if tmpItems[2] == "y":
                                                tmpTaskSpec.status = oldStatus
                                                tmpTaskSpec.forceUpdate("oldStatus")
                                                updateTaskStatus = False
                                    if commandStr == "reassign":
                                        tmpTaskSpec.forceUpdate("errorDialog")
                                    if commandStr == "finish":
                                        # update datasets
                                        tmpLog.info("updating datasets to finish")
                                        tmpStat = self.taskBufferIF.updateDatasetsToFinishTask_JEDI(jediTaskID, self.pid)
                                        if not tmpStat:
                                            tmpLog.info("wait until datasets are updated to finish")
                                        # ignore failGoalUnreached when manually finished
                                        tmpStat, taskSpec = self.taskBufferIF.getTaskWithID_JEDI(jediTaskID)
                                        tmpTaskSpec.splitRule = taskSpec.splitRule
                                        tmpTaskSpec.unsetFailGoalUnreached()
                                    if updateTaskStatus:
                                        tmpTaskSpec.status = JediTaskSpec.commandStatusMap()[commandStr]["done"]
                                    tmpMsg = f"set task_status={tmpTaskSpec.status}"
                                    tmpLog.sendMsg(tmpMsg, self.msgType)
                                    tmpLog.info(tmpMsg)
                                    tmpRet = self.taskBufferIF.updateTask_JEDI(tmpTaskSpec, {"jediTaskID": jediTaskID}, setOldModTime=True)
                                    tmpLog.info(f"done with {str(tmpRet)}")
                                    break
                                else:
                                    # kill only in the first loop
                                    if iLoop > 0:
                                        break
                                    # wait or kill jobs
                                    if commentStr and "soft finish" in commentStr:
                                        queuedPandaIDs = self.taskBufferIF.getQueuedPandaIDsWithTask_JEDI(jediTaskID)
                                        tmpMsg = f"trying to kill {len(queuedPandaIDs)} queued jobs for soft finish"
                                        tmpLog.info(tmpMsg)
                                        tmpRet = self.taskBufferIF.killJobs(queuedPandaIDs, commentStr, "52", True)
                                        tmpMsg = f"wating {len(pandaIDs)} jobs for soft finish"
                                        tmpLog.info(tmpMsg)
                                        tmpRet = True
                                        tmpLog.info(f"done with {str(tmpRet)}")
                                        break
                                    else:
                                        tmpMsg = f"trying to kill {len(pandaIDs)} jobs"
                                        tmpLog.info(tmpMsg)
                                        tmpLog.sendMsg(tmpMsg, self.msgType)
                                        if commandStr in ["finish"]:
                                            # force kill
                                            tmpRet = self.taskBufferIF.killJobs(pandaIDs, commentStr, "52", True)
                                        elif commandStr in ["reassign"]:
                                            # force kill
                                            tmpRet = self.taskBufferIF.killJobs(pandaIDs, commentStr, "51", True)
                                        else:
                                            # normal kill
                                            tmpRet = self.taskBufferIF.killJobs(pandaIDs, commentStr, "50", True)
                                        tmpLog.info(f"done with {str(tmpRet)}")
                    elif commandStr in ["retry", "incexec"]:
                        tmpMsg = f"executing {commandStr}"
                        tmpLog.info(tmpMsg)
                        tmpLog.sendMsg(tmpMsg, self.msgType)
                        # change task params for incexec
                        if commandStr == "incexec":
                            try:
                                # read task params
                                taskParam = self.taskBufferIF.getTaskParamsWithID_JEDI(jediTaskID)
                                taskParamMap = RefinerUtils.decodeJSON(taskParam)
                                # remove some params
                                for newKey in ["nFiles", "fixedSandbox"]:
                                    try:
                                        del taskParamMap[newKey]
                                    except Exception:
                                        pass
                                # convert new params
                                newParamMap = RefinerUtils.decodeJSON(commentStr)
                                # change params
                                for newKey, newVal in newParamMap.items():
                                    if newVal is None:
                                        # delete
                                        if newKey in taskParamMap:
                                            del taskParamMap[newKey]
                                    else:
                                        # change
                                        taskParamMap[newKey] = newVal
                                # overwrite sandbox
                                if "fixedSandbox" in taskParamMap:
                                    # noBuild
                                    for tmpParam in taskParamMap["jobParameters"]:
                                        if tmpParam["type"] == "constant" and re.search("^-a [^ ]+$", tmpParam["value"]) is not None:
                                            tmpParam["value"] = f"-a {taskParamMap['fixedSandbox']}"
                                    # build
                                    if "buildSpec" in taskParamMap:
                                        taskParamMap["buildSpec"]["archiveName"] = taskParamMap["fixedSandbox"]
                                    # merge
                                    if "mergeSpec" in taskParamMap:
                                        taskParamMap["mergeSpec"]["jobParameters"] = re.sub(
                                            "-a [^ ]+", f"-a {taskParamMap['fixedSandbox']}", taskParamMap["mergeSpec"]["jobParameters"]
                                        )
                                # encode new param
                                strTaskParams = RefinerUtils.encodeJSON(taskParamMap)
                                tmpRet = self.taskBufferIF.updateTaskParams_JEDI(jediTaskID, strTaskParams)
                                if tmpRet is not True:
                                    tmpLog.error("failed to update task params")
                                    continue
                            except Exception as e:
                                tmpLog.error(f"failed to change task params with {str(e)} {traceback.format_exc()}")
                                continue
                        # retry child tasks
                        if "sole " in commentStr:
                            retryChildTasks = False
                        else:
                            retryChildTasks = True
                        # discard events
                        if "discard " in commentStr:
                            discardEvents = True
                        else:
                            discardEvents = False
                        # release un-staged files
                        if "staged " in commentStr:
                            releaseUnstaged = True
                        else:
                            releaseUnstaged = False
                        tmpRet, newTaskStatus = self.taskBufferIF.retryTask_JEDI(
                            jediTaskID, commandStr, retryChildTasks=retryChildTasks, discardEvents=discardEvents, release_unstaged=releaseUnstaged
                        )
                        if tmpRet is True:
                            tmpMsg = f"set task_status={newTaskStatus}"
                            tmpLog.sendMsg(tmpMsg, self.msgType)
                            tmpLog.info(tmpMsg)
                        tmpLog.info(f"done with {tmpRet}")
                    else:
                        tmpLog.error("unknown command")
            except Exception as e:
                errStr = f"{self.__class__.__name__} failed in runImpl() with {str(e)} {traceback.format_exc()} "
                logger.error(errStr)


# launch


def launcher(commuChannel, taskBufferIF, ddmIF, vos=None, prodSourceLabels=None):
    p = TaskCommando(commuChannel, taskBufferIF, ddmIF, vos, prodSourceLabels)
    p.start()
